# -*- coding: utf-8 -*-
"""
SeeAlso:
    utool._internal.util_importer

    TODO: http://code.activestate.com/recipes/473888-lazy-module-imports/
    https://pypi.python.org/pypi/zope.deferredimport/3.5.2
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from utool import util_inject
# from utool import util_arg
import sys
print, rrr, profile = util_inject.inject2(__name__)


# def dynamic_import(modname, submod):
#     """
#     CommandLine:
#         python -m utool.util_import dynamic_import opengm.opengmcore _opengmcore

#     Tutorial:
#         ### Instead of writing:

#         from {your_submodule} import *
#         ### OR
#         from {your_module}.{your_submodule} import *

#         ### You can use utool's dynamic importer:

#         if True:
#             # To Autogenerate Run
#             '''
#             python -c "import {your_module}" --print-{your_module}-init
#             python -c "import {your_module}" --update-{your_module}-init
#             '''
#             import utool as ut
#             ut.dynamic_import(__name__, '{your_submodule}')
#         else:
#             # <AUTOGEN_INIT>
#             pass
#             # </AUTOGEN_INIT>

#        ### This will do the same thing, but you can "freeze" your module
#        ### and autogenerate what import * would have done. This helps
#        ### with static analysis and overall code readability.

#     Ignore:
#         modname = 'opengm.opengmcore'
#         submod = '_opengmcore'
#     """
#     from utool._internal import util_importer
#     if isinstance(submod, list):
#         import_tuples = submod
#     else:
#         import_tuples = [(submod, None)]
#     import_execstr, initstr = util_importer.dynamic_import(
#         modname, import_tuples, check_not_imported=False,
#         return_initstr=True)
#     return initstr


def import_star(modname, parent=None):
    """
    Args:
        modname (str): module name

    Tutorial:
        Replacement for
        from modname import *

        Usage is like this
        globals().update(ut.import_star('<module>'))
        OR
        ut.import_star('<module>', __name__)

    Ignore:
        >>> from utool.util_import import *  # NOQA
        modname = 'opengm'
        submod = None
        parent = __name__
    """
    import six
    from os.path import dirname
    if parent is None:
        parent_module = None
    else:
        parent_module = sys.modules[parent]
        if isinstance(parent, six.string_types):
            parent_module = sys.modules[parent]
        else:
            raise ValueError('parent must be the module __name__ attribute')

    try:
        module = sys.modules[modname]
    except KeyError:
        try:
            module = __import__(modname, {}, {}, fromlist=[], level=0)
        except ImportError:
            if parent_module is None:
                print('Maybe try specifying parent?')
                raise
            # Inject into the parent if given
            # Temporilly put this module dir in the pythonpath to simulate
            # relative imports
            relative_to = dirname(parent_module.__file__)
            sys.path.append(relative_to)
            try:
                module = __import__(modname, {}, {}, fromlist=[], level=0)
            except Exception:
                raise
            finally:
                sys.path.pop()
    # get public attributes
    module_attrs = [attr for attr in dir(module) if not attr.startswith('_')]
    module_vars = {attr: getattr(module, attr) for attr in module_attrs}
    if parent is not None:
        # Inject into the parent if given
        if isinstance(parent, six.string_types):
            parent_module = sys.modules[parent]
            for attr, var in module_vars.items():
                setattr(parent_module, attr, var)
        else:
            raise ValueError('parent must be the module __name__ attribute')
    # return the module dictionary
    return module_vars


def import_star_execstr(modname, parent=None):
    """
    print(ut.import_star_execstr('opengm.inference'))
    """
    from utool import util_str
    module_vars = import_star(modname, parent=parent)
    fromlist_str = ', '.join(sorted(module_vars.keys()))
    fromimport_prefix = 'from {modname} import ('.format(modname=modname)

    newline_prefix = (' ' * len(fromimport_prefix))
    if fromlist_str:
        rawstr = fromimport_prefix + fromlist_str + ',)'
    else:
        rawstr = ''
    textwidth = 79 - 4
    fromimport_str = util_str.pack_into(rawstr, textwidth=textwidth,
                                        newline_prefix=newline_prefix,
                                        break_words=False)

    # fromimport_str = ut.autopep8_format(fromimport_str, ignore={})
    return fromimport_str


# TODO import_freezestar


def possible_import_patterns(modname):
    r"""
    does not support from x import *
    does not support from x import z, y

    Example:
        >>> # DISABLE_DOCTEST
        >>> import utool as ut
        >>> modname = 'package.submod.submod2.module'
        >>> result = ut.repr3(ut.possible_import_patterns(modname))
        >>> print(result)
        [
            'import\\spackage.submod.submod2.module',
            'from\\spackage\\.submod\\.submod2\\simportmodule',
        ]
    """
    # common regexes
    WS = r'\s'
    import_ = 'import'
    from_ = 'from'
    dot_ = r'\.'
    patterns = [import_ + WS + modname]
    if '.' in modname:
        parts = modname.split('.')
        modpart = dot_.join(parts[0:-1])
        imppart = parts[-1]
        patterns += [from_ + WS + modpart + WS + import_ + imppart]
    NONSTANDARD = False
    if NONSTANDARD:
        if '.' in modname:
            for i in range(1, len(parts) - 1):
                modpart = '.'.join(parts[i:-1])
                imppart = parts[-1]
                patterns += [from_ + WS + modpart + WS + import_ + imppart]
            imppart = parts[-1]
            patterns += [import_ + WS + imppart]
    return patterns


def package_contents(package, with_pkg=False, with_mod=True, ignore_prefix=[],
                     ignore_suffix=[]):
    r"""
    References:
        http://stackoverflow.com/questions/1707709/list-all-the-modules-that-are-part-of-a-python-package

    Args:
        package (?):
        with_pkg (bool): (default = False)
        with_mod (bool): (default = True)

    CommandLine:
        python -m utool.util_import --exec-package_contents

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_import import *  # NOQA
        >>> import utool as ut
        >>> import ibeis
        >>> package = ibeis
        >>> ignore_prefix = ['ibeis.tests', 'ibeis.control.__SQLITE3__',
        >>>                  '_autogen_explicit_controller']
        >>> ignore_suffix = ['_grave']
        >>> with_pkg = False
        >>> with_mod = True
        >>> result = package_contents(package, with_pkg, with_mod,
        >>>                           ignore_prefix, ignore_suffix)
        >>> print(ut.repr2(result))
    """
    import pkgutil
    if not hasattr(package, '__path__'):
        return [package.__name__]
    #    pass
    print('package = %r' % (package,))
    walker = pkgutil.walk_packages(package.__path__,
                                   prefix=package.__name__ + '.',
                                   onerror=lambda x: None)
    module_list = []
    for importer, modname, ispkg in walker:
        if any(modname.startswith(prefix) for prefix in ignore_prefix):
            continue
        if any(modname.endswith(suffix) for suffix in ignore_suffix):
            continue
        if not ispkg and with_mod:
            module_list.append(modname)
        if ispkg and with_pkg:
            module_list.append(modname)
    return module_list


def get_modpath_from_modname(modname, prefer_pkg=False, prefer_main=False):
    """
    Same as get_modpath but doesnt import directly

    SeeAlso:
        get_modpath
    """
    from os.path import dirname, basename, join, exists
    initname = '__init__.py'
    mainname = '__main__.py'
    if modname in sys.modules:
        modpath = sys.modules[modname].__file__.replace('.pyc', '.py')
    else:
        import pkgutil
        loader = pkgutil.find_loader(modname)
        modpath = loader.filename.replace('.pyc', '.py')
        if '.' not in basename(modpath):
            modpath = join(modpath, initname)
    if prefer_pkg:
        if modpath.endswith(initname) or modpath.endswith(mainname):
            modpath = dirname(modpath)
    if prefer_main:
        if modpath.endswith(initname):
            main_modpath = modpath[:-len(initname)] + mainname
            if exists(main_modpath):
                modpath = main_modpath
    return modpath


def check_module_installed(modname):
    """
    Check if a python module is installed without attempting to import it.
    Note, that if ``modname`` indicates a child module, the parent module is
    always loaded.

    Args:
        modname (str):  module name

    Returns:
        bool: found

    References:
        http://stackoverflow.com/questions/14050281/module-exists-without-importing

    CommandLine:
        python -m utool.util_import check_module_installed --show --verbimp --modname=this
        python -m utool.util_import check_module_installed --show --verbimp --modname=ibeis.scripts.iccv

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_import import *  # NOQA
        >>> import utool as ut
        >>> modname = ut.get_argval('--modname', default='this')
        >>> is_installed = check_module_installed(modname)
        >>> is_imported = modname in sys.modules
        >>> print('module(%r).is_installed = %r' % (modname, is_installed))
        >>> print('module(%r).is_imported = %r' % (modname, is_imported))
        >>> assert 'this' not in sys.modules, 'module(this) should not have ever been imported'
    """
    import pkgutil
    if '.' in modname:
        # Prevent explicit import if possible
        parts = modname.split('.')
        base = parts[0]
        submods = parts[1:]
        loader = pkgutil.find_loader(base)
        if loader is not None:
            # TODO: check to see if path to the submod exists
            submods
            return True
    loader = pkgutil.find_loader(modname)
    is_installed = loader is not None
    return is_installed


def import_modname(modname):
    r"""
    Args:
        modname (str):  module name

    Returns:
        module: module

    CommandLine:
        python -m utool.util_import --test-import_modname

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_import import *  # NOQA
        >>> modname_list = [
        >>>     'utool',
        >>>     'utool._internal',
        >>>     'utool._internal.meta_util_six',
        >>>     'utool.util_path',
        >>>     #'utool.util_path.checkpath',
        >>> ]
        >>> modules = [import_modname(modname) for modname in modname_list]
        >>> result = ([m.__name__ for m in modules])
        >>> assert result == modname_list
    """
    # The __import__ statment is weird
    if util_inject.PRINT_INJECT_ORDER:
        if modname not in sys.modules:
            util_inject.noinject(modname, N=2, via='ut.import_modname')
    if '.' in modname:
        fromlist = modname.split('.')[-1]
        fromlist_ = list(map(str, fromlist))  # needs to be ascii for python2.7
        module = __import__(modname, {}, {}, fromlist_, 0)
    else:
        module = __import__(modname, {}, {}, [], 0)
    return module


def tryimport(modname, pipiname=None, ensure=False):
    """
    CommandLine:
        python -m utool.util_import --test-tryimport

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> modname = 'pyfiglet'
        >>> pipiname = 'git+https://github.com/pwaller/pyfiglet'
        >>> pyfiglet = ut.tryimport(modname, pipiname)
        >>> assert pyfiglet is None or isinstance(pyfiglet, types.ModuleType), 'unknown error'

    Example2:
        >>> # UNSTABLE_DOCTEST
        >>> # disabled because not everyone has access to being a super user
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> modname = 'lru'
        >>> pipiname = 'git+https://github.com/amitdev/lru-dict'
        >>> lru = ut.tryimport(modname, pipiname, ensure=True)
        >>> assert isinstance(lru, types.ModuleType), 'did not ensure lru'
    """
    if pipiname is None:
        pipiname = modname
    try:
        if util_inject.PRINT_INJECT_ORDER:
            if modname not in sys.modules:
                util_inject.noinject(modname, N=2, via='ut.tryimport')
        module = __import__(modname)
        return module
    except ImportError as ex:
        import utool as ut
        base_pipcmd = 'pip install %s' % pipiname
        sudo  = not ut.WIN32 and not ut.in_virtual_env()
        if sudo:
            pipcmd = 'sudo ' + base_pipcmd
        else:
            pipcmd = base_pipcmd
        msg = 'unable to find module %s. Please install: %s' % ((modname), (pipcmd))
        print(msg)
        ut.printex(ex, msg, iswarning=True)
        if ensure:
            raise AssertionError('Ensure is dangerous behavior and is is no longer supported.')
            #raise NotImplementedError('not ensuring')
            ut.cmd(base_pipcmd, sudo=sudo)
            module = tryimport(modname, pipiname, ensure=False)
            if module is None:
                raise AssertionError('Cannot ensure modname=%r please install using %r'  % (modname, pipcmd))
            return module
        return None


# lazy_module_attrs =  ['_modname', '_module', '_load_module']


# class LazyModule(object):
#     """
#     Waits to import the module until it is actually used.
#     Caveat: there is no access to module attributes used
#         in ``lazy_module_attrs``

#     CommandLine:
#         python -m utool.util_import --test-LazyModule

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from utool.util_import import *  # NOQA
#         >>> import sys
#         >>> assert 'this' not in sys.modules,  'this was imported before test start'
#         >>> this = LazyModule('this')
#         >>> assert 'this' not in sys.modules,  'this should not have been imported yet'
#         >>> assert this.i == 25
#         >>> assert 'this' in sys.modules,  'this should now be imported'
#         >>> print(this)
#     """
#     def __init__(self, modname):
#         r"""
#         Args:
#             modname (str):  module name
#         """
#         self._modname = modname
#         self._module = None

#     def _load_module(self):
#         if self._module is None:
#             if util_arg.VERBOSE:
#                 print('lazy loading module module')
#             self._module =  __import__(self._modname, globals(), locals(), fromlist=[], level=0)

#     def __str__(self):
#         return 'LazyModule(%s)' % (self._modname,)

#     def __dir__(self):
#         self._load_module()
#         return dir(self._module)

#     def __getattr__(self, item):
#         if item in lazy_module_attrs:
#             return super(LazyModule, self).__getattr__(item)
#         self._load_module()
#         return getattr(self._module, item)

#     def __setattr__(self, item, value):
#         if item in lazy_module_attrs:
#             return super(LazyModule, self).__setattr__(item, value)
#         self._load_module()
#         setattr(self._module, item, value)


def import_module_from_fpath(module_fpath):
    r""" imports module from a file path

    Args:
        module_fpath (str):

    Returns:
        module: module

    CommandLine:
        python -m utool.util_import --test-import_module_from_fpath

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_import import *  # NOQA
        >>> import utool
        >>> module_fpath = utool.__file__
        >>> module = import_module_from_fpath(module_fpath)
        >>> result = ('module = %s' % (str(module),))
        >>> print(result)


    Ignore:
        import shutil
        import ubelt as ub
        test_root = ub.ensure_app_cache_dir('test_fpath_import')
        # Clear the directory
        shutil.rmtree(test_root)
        test_root = ub.ensure_app_cache_dir('test_fpath_import')

        # -----
        # Define two temporary modules with the same name that are not in sys.path
        import sys, os, os.path
        from os.path import join

        # Even though they have the same name they have different values
        mod1_fpath = ub.ensuredir((test_root, 'path1', 'testmod'))
        ub.writeto(join(mod1_fpath, '__init__.py'), 'version = 1\nfrom . import sibling\na1 = 1')
        ub.writeto(join(mod1_fpath, 'sibling.py'), 'spam = \"ham\"\nb1 = 2')

        # Even though they have the same name they have different values
        mod2_fpath = ub.ensuredir((test_root, 'path2', 'testmod'))
        ub.writeto(join(mod2_fpath, '__init__.py'), 'version = 2\nfrom . import sibling\na2 = 3')
        ub.writeto(join(mod2_fpath, 'sibling.py'), 'spam = \"jam\"\nb2 = 4')

        # -----
        # Neither module should be importable through the normal mechanism
        try:
            import testmod
            assert False, 'should fail'
        except ImportError as ex:
            pass

        mod1 = ut.import_module_from_fpath(mod1_fpath)
        print('mod1.version = {!r}'.format(mod1.version))
        print('mod1.version = {!r}'.format(mod1.version))
        print(mod1.version == 1, 'mod1 version is 1')
        print('mod1.a1 = {!r}'.format(mod1.a1))

        mod2 = ut.import_module_from_fpath(mod2_fpath)
        print('mod2.version = {!r}'.format(mod2.version))
        print(mod2.version == 2, 'mod2 version is 2')
        print('mod2.a2 = {!r}'.format(mod1.a2))

        # BUT Notice how mod1 is mod2
        print(mod1 is mod2)

        # mod1 has attributes from mod1 and mod2
        print('mod1.a1 = {!r}'.format(mod1.a1))
        print('mod1.a2 = {!r}'.format(mod1.a2))
        print('mod2.a1 = {!r}'.format(mod2.a1))
        print('mod2.a2 = {!r}'.format(mod2.a2))

        # Both are version 2
        print('mod1.version = {!r}'.format(mod1.version))
        print('mod2.version = {!r}'.format(mod2.version))

        # However sibling always remains at version1 (ham)
        print('mod2.sibling.spam = {!r}'.format(mod2.sibling.spam))

        # now importing testmod works because it reads from sys.modules
        import testmod

        # reloading mod1 overwrites attrs again
        mod1 = ut.import_module_from_fpath(mod1_fpath)

        # Removing both from sys.modules
        del sys.modules['testmod']
        del sys.modules['testmod.sibling']
        mod2 = ut.import_module_from_fpath(mod2_fpath)

        print(not hasattr(mod2, 'a1'),
            'mod2 no longer has a1 and it reloads itself correctly')


        # -------

        del sys.modules['testmod']
        del sys.modules['testmod.sibling']
        mod1 = ut.import_module_from_fpath(mod1_fpath)


        # third test
        mod3_fpath = ub.ensuredir((test_root, 'path3', 'testmod'))
        ub.writeto(join(mod3_fpath, '__init__.py'), 'version = 3')

        module_fpath = mod3_fpath
        modname = 'testmod'

        # third test
        mod4_fpath = ub.ensuredir((test_root, 'path3', 'novelmod'))
        ub.writeto(join(mod4_fpath, '__init__.py'), 'version = 4')

    """
    from os.path import basename, splitext, isdir, join, exists, dirname, split
    import platform
    if isdir(module_fpath):
        module_fpath = join(module_fpath, '__init__.py')
    print('module_fpath = {!r}'.format(module_fpath))
    if not exists(module_fpath):
        raise ImportError('module_fpath={!r} does not exist'.format(
            module_fpath))
    python_version = platform.python_version()
    modname = splitext(basename(module_fpath))[0]
    if modname == '__init__':
        modname = split(dirname(module_fpath))[1]
    if util_inject.PRINT_INJECT_ORDER:
        if modname not in sys.argv:
            util_inject.noinject(modname, N=2, via='ut.import_module_from_fpath')
    if python_version.startswith('2.7'):
        import imp
        module = imp.load_source(modname, module_fpath)
    elif python_version.startswith('3'):
        import importlib.machinery
        loader = importlib.machinery.SourceFileLoader(modname, module_fpath)
        module = loader.load_module()
        # module = loader.exec_module(modname)
    else:
        raise AssertionError('invalid python version={!r}'.format(
            python_version))
    return module


#modname = 'theano'
#theano = LazyModule(modname)
if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_import
        python -m utool.util_import --allexamples
        python -m utool.util_import --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import functools
import six
import codecs

if not six.PY2:
    unicode = str

if six.PY2:
    import types
    __STR__ = unicode

    # TODO: use six.text_type
    def get_funcname(func):
        """
        Weird behavior for classes
        I dont know why this returns type / None
        import lasagne
        lasagne.layers.InputLayer
        lasagne.layers.InputLayer.__module__
        lasagne.layers.InputLayer.__class__.__name__ == 'type'
        lasagne.layers.InputLayer.__class__ is type
        wtf
        """
        try:
            return getattr(func, 'func_name')
        except AttributeError:
            builtin_function_name_dict = {
                len:    'len',
                zip:    'zip',
                range:  'range',
                map:    'map',
                type:   'type',
            }
            if func in builtin_function_name_dict:
                return builtin_function_name_dict[func]
            elif isinstance(func, functools.partial):
                return get_funcname(func.func)
            elif isinstance(func, six.class_types):
                return func.__name__
                #return str(func).replace('<class \'', '').replace('\'>', '')
            elif hasattr(func, '__class__'):
                return func.__class__.__name__
            else:
                print('Error inspecting func type')
                print(type(func))
                raise
        #except Exception as ex:
        #    import utool as ut
        #    with ut.embed_on_exception_context:
        #        raise

    def set_funcname(func, newname):
        return setattr(func, 'func_name', str(newname))
    #
    def get_funcglobals(func):
        """
        TODO: change all of these to work for classes and change the name to
        reflect that.
        """
        return getattr(func, 'func_globals')
        # try:
        #     return getattr(func, 'func_globals')
        # except AttributeError:
        #     if hasattr(func, '__class__'):
        #         # mod = sys.modules[func.__module__]
        #         # return {k: getattr(mod, k) for k in dir(mod)}
        #         return {}
        #     else:
        #         raise
    #
    def get_funcdoc(func):
        return getattr(func, 'func_doc')
    #
    def set_funcdoc(func, newdoc):
        return setattr(func, 'func_doc', newdoc)

    def get_funccode(func):
        return getattr(func, 'func_code')
    IntType  = types.IntType
    LongType = types.LongType
    BooleanType = types.BooleanType
    FloatType = types.FloatType
elif six.PY3:
    import types

    __STR__ = str
    IntType  = int
    LongType = int
    BooleanType = bool
    FloatType = float
    def get_funcname(func):
        try:
            return getattr(func, '__name__')
        except AttributeError as original:
            if isinstance(func, functools.partial):
                return get_funcname(func.func)
            if isinstance(func, types.BuiltinFunctionType):
                # for cv2.imread
                #return str(cv2.imread).replace('>', '').replace('<built-in function', '')
                return str(func).replace('<built-in function', '<')
            else:
                try:
                    return str(getattr(func, '__class__')).strip("<class '").strip("'>").split('.')[-1]
                except Exception:
                    raise original
    def set_funcname(func, newname):
        return setattr(func, '__name__', newname)
    #
    def get_funcglobals(func):
        return getattr(func, '__globals__')
    #
    def get_funcdoc(func):
        return getattr(func, '__doc__')
    def set_funcdoc(func, newdoc):
        return setattr(func, '__doc__', newdoc)

    def get_funccode(func):
        return getattr(func, '__code__')
else:
    raise AssertionError('python4 ?!!')


def ensure_unicode(str_):
    """
    TODO:
        rob gp "isinstance\\(.*\\\\bstr\\\\b\\)"
    """
    if isinstance(str_, __STR__):
        return str_
    else:
        try:
            return __STR__(str_)
        except UnicodeDecodeError:
            if str_.startswith(codecs.BOM_UTF8):
                # Can safely remove the utf8 marker
                # http://stackoverflow.com/questions/12561063/python-extract-data-from-file
                str_ = str_[len(codecs.BOM_UTF8):]
            return str_.decode('utf-8')
    #if not isinstance(str_, __STR__) and is_byte_encoded_unicode(str_):
    #    return str_.decode('utf-8')
    #else:
    #    return __STR__(str_)



# -*- coding: utf-8 -*-
#flake8:noqa
"""
This file is broken
"""
from __future__ import absolute_import, division, print_function
# from utool import regex


regex_on = r'\(^ \)\([^ ]*rrr.*\)'
repl_off = r'\1pass  # UTOOL_COMMENT \2'


regex_off = r'\(^ \)\(# UTOOL_COMMENT\)\([^ ]*rrr.*\)'
repl_on = r'\1\3'



# -*- coding: utf-8 -*-
"""
Helpers for tests

NOTE:
    DEPRICATE THIS IN FAVOR OF pytest and xdoctest

This module contains a more sane reimplementation of doctest functionality.
(I.E.  asserts work and you don't have to worry about stdout mucking things up)
The code isn't super clean though due to time constriaints.  Many functions
probably belong elsewhere and the parsers need a big cleanup.

TODO:
    * report the line of the doctest in the file when reporting errors as well as
     the relative line

    * restructure so there is a test collection step, a filtering step, and an
      execution step

    * Fix finding tests when running with @profile
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import inspect
import types
import sys
from collections import namedtuple
from os.path import basename
from utool import util_arg
from utool import util_time
from utool import util_inject
from utool import util_dbg
from utool import util_dev
from utool._internal.meta_util_six import get_funcname
print, rrr, profile = util_inject.inject2(__name__)


VERBOSE_TEST = util_arg.get_module_verbosity_flags('test')[0]

DEBUG_SRC = not util_arg.get_argflag('--nodbgsrc')
PRINT_SRC = util_arg.get_argflag(
    ('--printsrc', '--src', '--show-src', '--showsrc'),
    help_='show docstring source when running tests')
PRINT_FACE = util_arg.get_argflag(('--printface', '--face'))
BIGFACE = util_arg.get_argflag('--bigface')
SYSEXIT_ON_FAIL = util_arg.get_argflag(
    ('--sysexitonfail', '--fastfail'),
    help_='Force testing harness to exit on first test failure')
VERBOSE_TIMER = not util_arg.get_argflag('--no-time-tests')
INDENT_TEST   = False

ModuleDoctestTup = namedtuple('ModuleDoctestTup', ('enabled_testtup_list',
                                                   'frame_fpath',
                                                   'all_testflags', 'module'))


class TestTuple(util_dev.NiceRepr):
    """
    Simple container for test objects to replace old tuple format
    exec mode specifies if the test is being run as a script
    """
    def __init__(self, name, num, src, want, flag, tags=None, frame_fpath=None,
                 mode=None, nametup=None, test_namespace=None, shortname=None,
                 total=None):
        self._name = name  # function / class / testable name
        self.num = num    # doctest index
        self.src = src    # doctest src
        self.want = want  # doctest required result (optional)
        self.flag = flag  # doctest commandline flags
        self.frame_fpath = frame_fpath  # parent file fpath
        self.mode = mode      # flags if running as script
        self.nametup = nametup
        self.total = total
        self.test_namespace = test_namespace
        self.shortname = shortname

        if tags is None and src is not None:
            # hack to parse tag from source
            tagline = src.split('\n')[0].strip()
            if tagline.startswith('#'):
                tagline = tagline[1:].strip()
                tagline = tagline.replace('_DOCTEST', '')
                tags = tagline.split(',')
        self.tags = tags

    @property
    def name(self):
        # Hack for namespaced name
        return self.nametup[0]

    @property
    def full_name(self):
        return self.modname + '.' + self.name

    @property
    def modname(self):
        import utool as ut
        return ut.get_modname_from_modpath(self.frame_fpath)

    @property
    def namespace_levels(self):
        if self.test_namespace:
            return [self.modename, self.test_namespace]
        else:
            return [self.modename]

    @property
    def namespace(self):
        return '.'.join(self.namespace_levels)

    @property
    def exec_mode(self):
        return self.mode == 'exec'

    def __nice__(self):
        tagstr = ' ' + ','.join(self.tags) if self.tags is not None else ''
        return (' '  + self.name + ':' + str(self.num) + ' /' +
                str(self.total) + ' ' + tagstr + ' in ' + self.modname)


HAPPY_FACE_BIG = r'''
               .-""""""-.
             .'          '.
            /   O      O   \
           :                :
           |                |
           '  ,          ,' :
            \  '-......-'  /
             '.          .'
               '-......-'
                   '''

SAD_FACE_BIG = r'''
               .-""""""-.
             .'          '.
            /   O      O   \
           :           `    :
           |                |
           :    .------.    :
            \  '        '  /
             '.          .'
               '-......-'
                  '''

HAPPY_FACE_SMALL = r'''
     .""".
    | o o |
    | \_/ |
     ' = '
    '''

SAD_FACE_SMALL = r'''
     .""".
    | . . |
    |  ~  |
     ' = '
    '''

if BIGFACE:
    HAPPY_FACE = HAPPY_FACE_BIG
    SAD_FACE = SAD_FACE_BIG
else:
    HAPPY_FACE = HAPPY_FACE_SMALL
    #SAD_FACE = SAD_FACE_BIG
    SAD_FACE = SAD_FACE_SMALL


def get_package_testables(module=None, **tagkw):
    r"""
    New command that should eventually be used intead of old stuff?

    Args:
        module_list (list): (default = None)
        test_flags (None): (default = None)

    CommandLine:
        python -m utool get_package_testables --show --mod ibeis
        python -m utool get_package_testables --show --mod utool --tags SCRIPT
        python -m utool get_package_testables --show --mod utool --tags ENABLE

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> has_any = ut.get_argval('--tags', type_=str, default=None)
        >>> module = ut.get_argval('--mod', default='utool')
        >>> test_tuples = get_package_testables(module, has_any=has_any)
        >>> result = ut.repr3(test_tuples)
        >>> print(result)
        >>> #print(ut.repr3(ut.list_getattr(test_tuples, 'tags')))
    """
    import utool as ut
    if isinstance(module, six.string_types):
        module = ut.import_modname(module)
    modname_list = ut.package_contents(module, ignore_prefix=[], ignore_suffix=[])
    module_list = []
    for modname in modname_list:
        try:
            module_list.append(ut.import_modname(modname))
        except Exception:
            pass

    test_tuples = []
    for module in module_list:
        old_testables = ut.get_module_doctest_tup(module=module,
                                                  needs_enable=False,
                                                  allexamples=True, verbose=False)
        test_tuples.extend(old_testables.enabled_testtup_list)
    if tagkw:
        tags_list = ut.list_getattr(test_tuples, 'tags')
        flags = ut.filterflags_general_tags(tags_list, **tagkw)
        test_tuples = ut.compress(test_tuples, flags)
    return test_tuples


def doctest_module_list(module_list):
    """
    Runs many module tests

    Entry point for batch run
    Depth 0)

    Ignore:
        :'<,'>!sort -n -k 2
    """
    import utool as ut
    nPass_list = []
    nTotal_list = []
    failed_cmds_list = []
    error_reports_list = []
    print('[util_test] Running doctests on module list')

    try:
        ut.write_to('timeings.txt', '\n\n --- begining doctest_module_list\n', mode='a')
    except IOError as ex:
        ut.printex(ex, '[util_test] IOWarning', iswarning=True)

    failed_doctest_fname = 'failed_doctests.txt'
    seen_ = set([])
    with open(failed_doctest_fname, 'a') as file_:
        file_.write('\n-------\n\n')
        file_.write(ut.get_timestamp(format_='printable') + '\n')
        file_.write('logfile (only present if logging) = %r\n' %
                    (ut.util_logging.get_current_log_fpath(),))
        testkw = dict(allexamples=True)
        with ut.Timer(verbose=False) as t:
            for module in module_list:
                (nPass, nTotal, failed_list, error_report_list) = ut.doctest_funcs(
                    module=module, seen_=seen_, **testkw)
                nPass_list.append(nPass)
                nTotal_list.append(nTotal)
                failed_cmds_list.append(failed_list)
                error_reports_list.append(error_report_list)
                # Write failed tests to disk
                for cmd in failed_list:
                    file_.write(cmd + '\n')
        total_time = t.ellapsed
        nPass = sum(nPass_list)
        nTotal = sum(nTotal_list)
        file_.write('PASSED %d / %d' % (nPass, nTotal))

    failed_cmd_list = ut.flatten(failed_cmds_list)
    error_report_list = ut.filter_Nones(ut.flatten(error_reports_list))
    if len(error_report_list) > 0:
        print('\nPrinting %d error reports' % (len(error_report_list),))
        for count, error_report in enumerate(error_report_list):
            print('\n=== Error Report %d / %d' % (count, len(error_report_list)))
            print(error_report)
        print('--- Done printing error reports ----')

    try:
        ut.write_to('timeings.txt', '\n\n --- finished doctest_module_list total_time=%.3fs\n' % (total_time), mode='a')
    except IOError as ex:
        ut.printex(ex, '[util_test] IOWarning', iswarning=True)

    print('')
    print('+========')
    print('| FINISHED TESTING %d MODULES' % (len(module_list),))
    print('| PASSED %d / %d' % (nPass, nTotal))
    print('L========')
    if len(failed_cmd_list) > 0:
        print('FAILED TESTS:')
        print('\n'.join(failed_cmd_list))
    return nPass, nTotal, failed_cmd_list


def doctest_funcs(testable_list=None, check_flags=True, module=None,
                  allexamples=None, needs_enable=None, strict=False,
                  verbose=True, return_error_report=True, seen_=None):
    r"""
    Main entry point into utools main module doctest harness
    Imports a module and checks flags for the function to run
    Depth 1)

    Args:
        testable_list (list):
        check_flags (bool): Force checking of the --test- and --exec- flags
        module (None):
        allexamples (None):
        needs_enable (None):

    Returns:
        tuple: (nPass, nTotal, failed_cmd_list)

    CommandLine:
        python -m ibeis.algo.preproc.preproc_chip --all-examples

    References:
        http://legacy.python.org/dev/peps/pep-0338/
        https://docs.python.org/2/library/runpy.html

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> testable_list = []
        >>> check_flags = True
        >>> module = None
        >>> allexamples = None
        >>> needs_enable = None
        >>> # careful might infinitely recurse
        >>> (nPass, nTotal) = doctest_funcs(testable_list, check_flags, module,
        ...                                 allexamples, needs_enable)
        >>> print((nPass, nTotal))
    """
    import multiprocessing
    import utool as ut  # NOQA
    #ut.start_logging()
    multiprocessing.freeze_support()  # just in case
    if ut.VERBOSE:
        print('[util_test] doctest_funcs')
    ut.inject_colored_exceptions()

    if (verbose or VERBOSE_TEST) and ut.NOT_QUIET:
        if VERBOSE_TEST:
            print('[util_test.doctest_funcs][DEPTH 1] doctest_funcs()')
        print('[util_test.doctest_funcs] Running doctest_funcs')
    if ut.is_developer():
        ut.change_term_title('DocTest ' + ' '.join(sys.argv))

    # PARSE OUT TESTABLE DOCTESTTUPS
    mod_doctest_tup = get_module_doctest_tup(
        testable_list, check_flags, module, allexamples, needs_enable, N=1,
        verbose=verbose)
    enabled_testtup_list, frame_fpath, all_testflags, module = mod_doctest_tup

    nPass = 0
    nFail = 0
    nTotal = len(enabled_testtup_list)

    #flags = [(tup.name, tup.num) in seen_ for tup in enabled_testtup_list]
    if seen_ is not None:
        flags = [tup.src not in seen_ for tup in enabled_testtup_list]
        enabled_testtup_list = ut.compress(enabled_testtup_list, flags)

    # Remove duplicate tests from previous parts of the batch run
    #print(sum(flags))

    EARLYEXIT = False
    if seen_ is not None:
        for tup in enabled_testtup_list:
            #seen_.add((tup.name, tup.num))
            seen_.add(tup.src)
    if EARLYEXIT:
        nPass = nTotal - sum(flags)
        if return_error_report:
            return (nPass, nTotal, [], [])
        else:
            return (nPass, nTotal, [])

    modname = ut.get_modname_from_modpath(frame_fpath)
    nTotal = len(enabled_testtup_list)

    # Run enabled examles
    failed_flag_list = []
    error_report_list = []
    if ut.get_argflag(('--edit-test-file', '--etf')):
        ut.editfile(frame_fpath)
    exec_mode = all([testtup.exec_mode for testtup in enabled_testtup_list])

    for testtup in enabled_testtup_list:
        name = testtup.name
        num  = testtup.num
        src  = testtup.src
        want = testtup.want
        flag = testtup.flag
        #if ut.is_developer():
        #    ut.change_term_title('DocTest ' + modname + ' ' + name)
        print('\n')
        fmtdict = dict(modname=modname, name=name, num=num)
        #      1          v12     v20       v30       v40       v50         v62
        print('+------------------------------------------------------------+')
        print('*  DOCTEST {modname:<20} {name:>26}:{num:d} '.format(**fmtdict))
        print('+------------------------------------------------------------+')

        if PRINT_SRC or VERBOSE_TEST:
            if ut.is_developer():
                print(ut.msgblock('EXEC SRC', ut.highlight_code(src), side='>>>'))
            else:
                print(ut.msgblock('EXEC SRC', src, side='>>>'))
        # Commented because it caused differences between
        # individual test runs and large test runs with ut
        # being imported
        # test_globals = module.__dict__.copy()
        test_globals = {}
        error_report = None
        try:
            testkw = dict(
                globals=test_globals,  # HACK
                want=want)
            assert testtup.frame_fpath == frame_fpath
            test_locals, error_report = ut.run_test(testtup, **testkw)
            pass_flag = (test_locals is not False)
            if pass_flag:
                if VERBOSE_TEST:
                    print('seems to pass')
                nPass += 1
            else:
                if VERBOSE_TEST:
                    print('raising failed exception')
                raise Exception('failed')
        except Exception:
            if VERBOSE_TEST:
                print('Seems to fail. ')
            nFail += 1
            failed_flag_list.append(flag)
            error_report_list.append(error_report)
            if strict or util_arg.SUPER_STRICT:
                raise
            else:
                if VERBOSE_TEST:
                    print('Silently Failing: '
                          'maybe adding the --super-strict flag would help debug?')
            pass
        except KeyboardInterrupt:
            print('[util_test] caught Ctrl+C')
        print('L_____________________________________________________________')
    #L__________________
    #+-------------------
    # Print Results
    if nTotal == 0 and not allexamples:
        valid_test_argflags = ['--allexamples'] + all_testflags
        modname = ut.get_modname_from_modpath(frame_fpath)
        # TODO: ensure that exename is in the PATH
        exename = basename(sys.executable)
        warning_msg = ut.codeblock(
            r'''
            [ut.doctest_funcs] No test flags specified
            Please choose one of the following flags or specify --enableall
            Valid test argflags:
            ''') + ut.indentjoin(valid_test_argflags, '\n %s -m %s ' % (exename, modname))
        # warning_msg = ut.indent(warning_msg, '[util_test.doctest_funcs]')
        ut.colorprint(warning_msg, 'yellow')

    if not exec_mode:
        print('+-------')
        print('| finished testing fpath=%r' % (frame_fpath,))
        print('| passed %d / %d' % (nPass, nTotal))
        print('L-------')
    failed_cmd_list = []
    if nFail > 0:
        #modname = module.__name__
        modname = ut.get_modname_from_modpath(frame_fpath)
        # TODO: ensure that exename is in the PATH
        exename = basename(sys.executable)
        failed_cmd_list = ['%s -m %s %s' % (exename, modname, flag_)
                            for flag_ in failed_flag_list]
        #failed_cmd_list = ['python %s %s' % (frame_fpath, flag_)
        #                    for flag_ in failed_flag_list]
        print('Failed sys.argv = %r' % (' '.join(sys.argv),))
        print('Failed Tests:')
        print('\n'.join(failed_cmd_list))
    #L__________________

    if ut.util_inject.PROFILING:
        ut.dump_profile_text()

    if return_error_report:
        return (nPass, nTotal, failed_cmd_list, error_report_list)
    else:
        return (nPass, nTotal, failed_cmd_list)


def run_test(func_or_testtup, *args, **kwargs):
    r"""
    Runs the test function. Prints a success / failure report.

    Args:
        func_or_testtup (func or tuple): function or doctest tuple
        args*: args to be forwarded to `func_or_testtup`
        kwargs*: keyword args to be forwarded to `func_or_testtup`
    """
    import utool as ut
    #func_is_testtup = isinstance(func_or_testtup, tuple)
    # NOTE: isinstance might not work here if ut.rrrr has been called
    func_is_testtup = isinstance(func_or_testtup, TestTuple)
    exec_mode = False
    dump_mode = False
    write_times = True
    if func_is_testtup:
        testtup = func_or_testtup
        src         = testtup.src
        funcname    = testtup.name
        frame_fpath = testtup.frame_fpath
        #(funcname, src, frame_fpath) = func_or_testtup
        exec_mode = testtup.exec_mode
        dump_mode = testtup.mode == 'dump'
    else:
        func_ = func_or_testtup
        funcname = get_funcname(func_)
        frame_fpath = ut.get_funcfpath(func_)
    upper_funcname = funcname.upper()
    if ut.VERBOSE:
        print('\n=============================')
        print('**[TEST.BEGIN] %s ' % (sys.executable))
        print('**[TEST.BEGIN] %s ' % (funcname,))

    verbose_timer = not exec_mode and VERBOSE_TIMER
    nocheckwant = True if exec_mode else None
    print_face = not exec_mode and PRINT_FACE
    error_report = None

    if dump_mode:
        print('testtup = %r' % (testtup,))
        print(ut.highlight_code(src))
        return None, None

    try:
        # RUN THE TEST WITH A TIMER
        with util_time.Timer(upper_funcname, verbose=verbose_timer) as timer:
            if func_is_testtup:
                test_locals = _exec_doctest(src, kwargs, nocheckwant)
            else:
                # TEST INPUT IS A LIVE PYTHON FUNCTION
                test_locals = func_(*args, **kwargs)
    except Exception as ex:
        import utool as ut
        # Get locals in the wrapped function
        ut.printex(ex, tb=True)
        error_report_lines = ['**[TEST.ERROR] %s -- FAILED:\n    type(ex)=%s' % (
            funcname, type(ex))]
        error_report_lines.append(ut.formatex(ex, tb=True))
        def print_report(msg):
            error_report_lines.append(msg)
            print(msg)
        print_report('\n=============================')
        print_report('**[TEST.FINISH] %s -- FAILED:\n    type(ex)=%s' % (funcname, type(ex)))
        exc_type, exc_value, tb = sys.exc_info()
        if PRINT_FACE:
            print_report(SAD_FACE)
        if func_is_testtup:
            print_report('Failed in module: %r' % frame_fpath)
            src_with_lineno = ut.number_text_lines(src)
            print_report(ut.msgblock('FAILED DOCTEST IN %s' % (funcname,), src_with_lineno))

        if util_arg.SUPER_STRICT:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            if not func_is_testtup:
                # Remove this function from stack strace
                # dont do this for execed code
                exc_traceback = exc_traceback.tb_next
                pass
            # FIXME: make a _internal/py2_syntax_funcs version of this
            # function as well as a python3 syntax version due to the
            # reraise syntax issue to avoid an extra frame in the stack
            six.reraise(exc_type, exc_value, exc_traceback)
        if SYSEXIT_ON_FAIL:
            print('[util_test] SYSEXIT_ON_FAIL = True')
            print('[util_test] exiting with sys.exit(1)')
            sys.exit(ut.EXIT_FAILURE)
        #raise
        error_report = '\n'.join(error_report_lines)
        return False, error_report
    else:
        # LOG PASSING TEST
        if not exec_mode:
            print('\n=============================')
            print('**[TEST.FINISH] %s -- SUCCESS' % (funcname,))
            if print_face:
                print(HAPPY_FACE)
            if write_times:
                timemsg = '%.4fs in %s %s\n' % (
                    timer.ellapsed, funcname, frame_fpath)
                try:
                    ut.write_to('timeings.txt', timemsg, mode='a')
                except IOError as ex:
                    ut.printex(ex, '[util_test] IOWarning', iswarning=True)
        # RETURN VALID TEST LOCALS
        return test_locals, error_report


def _exec_doctest(src, kwargs, nocheckwant=None):
    """
    Helper for run_test

    block of code that runs doctest and was too big to be in run_test
    """
    # TEST INPUT IS PYTHON CODE TEXT
    #test_locals = {}
    test_globals = kwargs.get('globals', {}).copy()
    want = kwargs.get('want', None)
    #test_globals['print'] = doctest_print
    # EXEC FUNC
    #six.exec_(src, test_globals, test_locals)  # adds stack to debug trace
    import utool as ut
    # TODO RECTIFY WITH TF
    if ut.get_argflag(('--cmd', '--embed')):
        src = _cmd_modify_src(src)
    code = compile(src, '<string>', 'exec')
    try:
        # IN EXEC CONTEXT THERE IS NO DIFF BETWEEN LOCAL / GLOBALS.  ONLY PASS
        # IN ONE DICT. OTHERWISE TREATED ODDLY
        # References: https://bugs.python.org/issue13557
        #exec(code, test_globals, test_locals)
        test_locals = test_globals
        exec(code, test_globals)
    except ExitTestException:
        print('Test exited before show')
        pass
    if nocheckwant is None:
        nocheckwant = util_arg.get_argflag(
            '--no-checkwant', help_='Turns off checking for results')
    if nocheckwant or want is None or want == '':
        if not nocheckwant:
            print('warning test does not want anything')
    else:
        if want.endswith('\n'):
            want = want[:-1]
        result = six.text_type(test_locals.get('result',
                                               'NO VARIABLE NAMED result'))
        if result != want:
            from utool import util_str
            msglines = []
            # Even if they arent exactly the same, the difference might just be
            # some whitespace characters. Ignore in this case.
            difftext = util_str.get_textdiff(want, result,
                                             ignore_whitespace=True)
            if difftext:
                if util_dbg.COLORED_EXCEPTIONS:
                    difftext = ut.color_diff_text(difftext)
                msglines += [
                    'DIFF/GOT/EXPECTED',
                    difftext,
                ]
                if VERBOSE_TEST:
                    msglines += [
                        'REPR_GOT: result=',
                        repr(result),
                        'REPR_EXPECTED: want=',
                        repr(want),
                    ]
                msglines += [
                    'STR_GOT: result=',
                    str(result),
                    'STR_EXPECTED: want=',
                    str(want),
                ]
                errmsg1 = '\n'.join(msglines)
                raise AssertionError('result != want\n' + errmsg1)
    return test_locals


def get_module_testlines(module_list, remove_pyc=True, verbose=True,
                         pythoncmd=None):
    """
    Builds test commands for autogen tests
    called by autogen test scripts
    """
    import utool as ut  # NOQA
    if pythoncmd is None:
        pythoncmd = sys.executable
        #'python'
    testcmd_list = []
    for module in module_list:
        mod_doctest_tup = get_module_doctest_tup(
            module=module, allexamples=True, verbose=verbose)
        (enabled_testtup_list, frame_fpath,
                 all_testflags, module_) = mod_doctest_tup
        for testtup in enabled_testtup_list:
            #testflag = testtup[-1]
            testflag = testtup.flag
            if remove_pyc:
                # FIXME python 3 __pycache__/*.pyc
                frame_fpath = frame_fpath.replace('.pyc', '.py')
            frame_rel_fpath = ut.get_relative_modpath(frame_fpath)
            testcmd = ' '.join((pythoncmd, frame_rel_fpath, testflag))
            testcmd_list.append(testcmd)
    return testcmd_list


def _docblock_tester1():
    """testting docblocks
        foo
    # Depth 5)
    call:
        foo
    the next call is:
        bar
    called by parse_doctest_from_docstr
    TODO: move to util_inspect
        test what happens with weird indentation
    bla

    Args:
        docstr (str): a documentation string formatted in google style.
            Remember indentation matters.
              whoo
            this
              .. params
        new (Optional[bool]): does a new style of parsing
        **kwargs:

    Returns:
        list: docstr_blocks tuples
            [(blockname, blockstr, offset)]

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> print('example1')

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> print('dummy')
    """


def _docblock_tester2():
    """
    Args:
    Returns:
    Example:
    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
    """


def _docblock_tester3():
    """
    foo
    Args:

    Returns:

    Example:

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
    """


def _docblock_tester4():
    """foobar

    Args:
        *args:
        **kwargs:

    Returns:
        bool

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
    """


def _test_docblock_parser():
    import utool as ut
    basis = {
        'n_args': [None, 0, 1],
        'n_return': [None, 0, 1, 2, 3],
        'spacing': [1, 2],
        # 'example_lines': [0, 1, 2],
        'doc_lines': [0, 1, 2],
        'doc_indent': [0, 1, 4],
    }
    for config in ut.all_dict_combinations(basis):
        print('---')
        print(config)
        print('=====')
        docstr =  _make_test_docstr(config)
        docparts = ut.parse_docblocks_from_docstr(docstr)
        print(docstr)
        print('=====')
        if config['n_args']:
            assert 'Args:' in ut.take_column(docparts, 0)
        if config['n_return']:
            assert 'Returns:' in ut.take_column(docparts, 0)


def _make_test_docstr(config):
    import utool as ut

    blocks = []

    docpart = '\n'.join(ut.lorium_ipsum().strip().split('\n')[0:config['doc_lines']])
    docpart = (' ' * config['doc_indent']) + docpart
    blocks.append(docpart)

    if config['n_args'] is not None:
        argheader = 'Args'
        argname_list = ut.chr_range(config['n_args'], base='a')
        argtype_list = ['bool'] * config['n_args']
        argdesc_list = [''] * config['n_args']
        arg_docstr = ut.make_args_docstr(argname_list, argtype_list,
                                         argdesc_list, ismethod=False)
        argsblock = ut.make_docstr_block(argheader, arg_docstr)
    else:
        argsblock = ''

    if config['n_return'] is not None:
        if config['n_return'] == 0:
            returnblock = 'Returns:'
        else:
            return_name = 'outvar'
            return_type = 'int'
            return_desc = '\n'.join(ut.lorium_ipsum().strip().split('\n')[0:config['n_return']])
            return_doctr = ut.make_returns_or_yeilds_docstr(return_type, return_name, return_desc)
            returnblock = ut.make_docstr_block('Returns', return_doctr)
        blocks.append(returnblock)

    blocks.append(argsblock)

    docstr = ('\n' * config['spacing']).join(blocks)
    return docstr


def parse_docblocks_from_docstr(docstr, offsets=False):
    """
    # Depth 5)
    called by parse_doctest_from_docstr

    TODO: move to util_inspect

    Args:
        docstr (str): a documentation string formatted in google style.

    Returns:
        list: docstr_blocks tuples
            [(blockname, blockstr)]

    CommandLine:
        python -m utool.util_tests parse_docblocks_from_docstr

    Doctest:
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> #func_or_class = ut.flatten2
        >>> #func_or_class = ut.list_depth
        >>> func_or_class = ut.iter_module_doctestable
        >>> func_or_class = ut.all_multi_paths
        >>> docstr = ut.get_docstr(func_or_class)
        >>> docstr_blocks = parse_docblocks_from_docstr(docstr)
        >>> result = str(ut.repr4(docstr_blocks))
        >>> print(result)
    """
    if docstr is None:
        return []
    # import parse
    import utool as ut
    import re
    # import itertools as it
    docstr = ut.ensure_unicode(docstr)
    docstr = ut.unindent(docstr)

    # Parse out initial documentation lines
    # Then parse out the blocked lines.
    docstr_lines = docstr.split('\n')
    line_indent = [ut.get_indentation(line) for line in docstr_lines]
    line_len = [len(line) for line in docstr_lines]

    # The first line may not have the correct indentation if it starts
    # right after the triple quotes. Adjust it in this case to ensure that
    # base indent is always 0
    adjusted = False
    is_nonzero = [len_ > 0 for len_ in line_len]
    if len(line_indent) >= 2:
        if line_len[0] != 0:
            indents = ut.compress(line_indent, is_nonzero)
            if len(indents) >= 2:
                indent_adjust = min(indents[1:])
                line_indent[0] += indent_adjust
                line_len[0] += indent_adjust
                docstr_lines[0] = (' ' * indent_adjust) + docstr_lines[0]
                adjusted = True
    if adjusted:
        # Redo prepreocessing, but this time on a rectified input
        docstr = ut.unindent('\n'.join(docstr_lines))
        docstr_lines = docstr.split('\n')
        line_indent = [ut.get_indentation(line) for line in docstr_lines]
        line_len = [len(line) for line in docstr_lines]

    indents = ut.compress(line_indent, is_nonzero)
    if len(indents) >= 1:
        if indents[0] != 0:
            print('ERROR IN PARSING')
            print('adjusted = %r' % (adjusted,))
            print(docstr)
            raise ValueError('Google Style Docstring Missformat')

    base_indent = 0
    # We will group lines by their indentation.
    # Rectify empty lines by giving them their parent's indentation.
    true_indent = []
    prev_indent = None
    for indent_, len_ in zip(line_indent, line_len):
        if len_ == 0:
            # Empty lines take on their parents indentation
            indent_ = prev_indent
        true_indent.append(indent_)
        prev_indent = indent_

    # tag_pattern = ut.regex_or(['Args:', 'Return:', 'CommandLine':])
    tag_pattern = '[^\\s]+: *$'

    group_id = 0
    prev_indent = 0
    group_list = []
    in_tag = False
    for line_num, (line, indent_) in enumerate(zip(docstr_lines, true_indent)):
        if re.match(tag_pattern, line):
            # Check if we can look ahead
            if line_num + 1 < len(docstr_lines):
                # A tag is only valid if its next line is properly indented,
                # empty, or is a tag itself.
                if (true_indent[line_num + 1] > base_indent or line_len[line_num + 1] == 0 or re.match(tag_pattern, docstr_lines[line_num + 1])):
                    group_id += 1
                    in_tag = True
            else:
                group_id += 1
                in_tag = True
        # Check if this line belongs to a new group
        elif in_tag and indent_ != prev_indent and indent_ == base_indent:
            group_id += 1
            in_tag = False
        group_list.append(group_id)
        prev_indent = indent_

    groups_ = ut.group_items(docstr_lines, group_list)
    groups = []
    line_offset = 0
    for k, lines in groups_.items():
        if len(lines) == 0 or (len(lines) == 1 and len(lines[0]) == 0):
            continue
        elif len(lines) >= 1 and re.match(tag_pattern, lines[0]):
            # An encoded google sub-block
            key = lines[0]
            val = lines[1:]
            subblock = ut.unindent('\n'.join(val))
        else:
            # A top level text documentation block
            key = '__DOC__'
            val = lines[:]
            subblock = '\n'.join(val)

        if offsets:
            groups.append((key, subblock, line_offset))
        else:
            groups.append((key, subblock))
        line_offset += len(lines)
    # Ensure that no keys are duplicated
    # try:
    #     assert len(ut.find_duplicate_items(ut.take_column(groups, 0))) == 0, (
    #         'Duplicate google docblock keys are not allowed')
    # except Exception as ex:
    #     ut.printex(ex, iswarning=True)
    return groups


def read_exampleblock(docblock):
    # if False:
    #     no longer needed with new parser
    #     import utool as ut
    #     nonheader_src = ut.unindent('\n'.join(docblock.splitlines()[1:]))
    nonheader_src = docblock
    nonheader_lines = nonheader_src.splitlines()
    reversed_src_lines = []
    reversed_want_lines = []
    finished_want = False

    # Read the example block backwards to get the want string
    # and then the rest should all be source
    for line in reversed(nonheader_lines):
        if not finished_want:
            if line.startswith('>>> ') or line.startswith('... '):
                finished_want = True
            else:
                reversed_want_lines.append(line)
                if len(line.strip()) == 0:
                    reversed_want_lines = []
                continue
        reversed_src_lines.append(line[4:])
    test_src = '\n'.join(reversed_src_lines[::-1])
    test_want = '\n'.join(reversed_want_lines[::-1])
    return test_src, test_want


def parse_doctest_from_docstr(docstr):
    r"""
    because doctest itself doesnt do what I want it to do

    CAREFUL, IF YOU GET BACK WRONG RESULTS MAKE SURE YOUR DOCSTR IS PREFFIXED
    WITH R

    CommandLine:
        python -m utool.util_tests --exec-parse_doctest_from_docstr

    Doctest:
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> func_or_class = parse_doctest_from_docstr
        >>> func_or_class = ut.list_depth
        >>> #func_or_class = ut.util_depricated.cartesian
        >>> func_or_class = ut.iter_module_doctestable
        >>> docstr = ut.get_docstr(func_or_class)
        >>> testsrc_list, testwant_list, testlinenum_list, func_lineno, docstr = get_doctest_examples(func_or_class)
        >>> print('\n\n'.join(testsrc_list))
        >>> assert len(testsrc_list) == len(testwant_list)
    """
    import utool as ut
    docstr_blocks = parse_docblocks_from_docstr(docstr, offsets=True)
    # print('docstr_blocks = %r' % (docstr_blocks,))

    example_docblocks = []
    example_setups = []
    grid_example_docblock = []
    grid_setups = []
    param_grids = None

    for header, docblock, line_offset in docstr_blocks:
        if header.startswith('Example') or header.startswith('Doctest'):
            example_docblocks.append((header, docblock, line_offset))

        if header.startswith('Setup'):
            setup_src = read_exampleblock(docblock)[0]
            example_setups.append(setup_src)

        if header.startswith('GridParams'):
            paramgrid_src = read_exampleblock(docblock)[0]
            globals_ = {}
            six.exec_('import utool as ut\n' + paramgrid_src, globals_)
            assert 'combos' in globals_, 'param grid must define combos'
            combos = globals_['combos']
            param_grids = [ut.execstr_dict(combo, explicit=True)
                           for combo in combos]

        if header.startswith('GridExample'):
            grid_example_docblock.append((header, docblock, line_offset))

        if header.startswith('GridSetup'):
            setup_src = read_exampleblock(docblock)[0]
            grid_setups.append(setup_src)

    assert len(example_setups) <= 1, 'cant have more than 1 setup. %d' % (
        len(example_setups))

    if example_setups and not grid_setups:
        grid_setups = example_setups

    testheader_list     = []
    testsrc_list        = []
    testwant_list       = []
    testlineoffset_list = []

    # Place grid tests first
    for header, docblock, line_offset in grid_example_docblock:
        test_src, test_want = read_exampleblock(docblock)
        assert len(grid_setups) <= 1, 'need one grid setup'
        if len(grid_setups):
            grid_setup = grid_setups[0]
        else:
            grid_setup = ''
        hack_show_request = False
        if 'ut.show_if_requested()' in test_src:
            hack_show_request = True
            test_src = test_src.replace('ut.show_if_requested()', '')  # Megahack
        full_grid_testsrc = '\n'.join(
            [grid_setup] + [
                '\n'.join([
                    pgrid, 'print(\'Grid %d\')' % (count,), test_src
                    if count < (len(param_grids) - 1) else test_src
                ])
                for count, pgrid in enumerate(param_grids)])
        if hack_show_request:
            full_grid_testsrc += '\n' + 'ut.show_if_requested()'

        testsrc_list.append(full_grid_testsrc)
        testheader_list.append(header)
        testwant_list.append(test_want)
        testlineoffset_list.append(line_offset)

    for header, docblock, line_offset in example_docblocks:
        test_src, test_want = read_exampleblock(docblock)
        if header.startswith('Doctest'):
            # Enable anything labeled as a doctest
            test_src = '# ENABLE_DOCTEST\n' + test_src
        if len(example_setups) == 0:
            full_testsrc = test_src
        elif len(example_setups) == 1:
            # Hack: append setups to all sources
            full_testsrc = '\n'.join([example_setups[0], test_src])
        else:
            assert False, 'more than 1 setup'
        testheader_list.append(header)
        testsrc_list.append(full_testsrc)
        testwant_list.append(test_want)
        testlineoffset_list.append(line_offset)

    return testheader_list, testsrc_list, testwant_list, testlineoffset_list


#@debug_decor
def get_doctest_examples(func_or_class, modpath=None):
    """
    get_doctest_examples

    Depth 3)
    called by get_module_doctest_tup

    Args:
        func_or_class (function)

    Returns:
        tuple (list, list): example_list, want_list

    CommandLine:
        python -m utool.util_tests --test-get_doctest_examples

    Doctest:
        >>> from utool.util_tests import *  # NOQA
        >>> func_or_class = get_doctest_examples
        >>> tup  = get_doctest_examples(func_or_class)
        >>> testsrc_list, testwant_list, testlinenum_list, func_lineno, docstr = tup
        >>> result = str(len(testsrc_list) + len(testwant_list))
        >>> print(testsrc_list)
        >>> print(testlinenum_list)
        >>> print(func_lineno)
        >>> print(testwant_list)
        >>> print(result)
        6

    Doctest:
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> func_or_class = ut.tryimport
        >>> tup = get_doctest_examples(func_or_class)
        >>> testsrc_list, testwant_list, testlinenum_list, func_lineno, docstr = tup
        >>> result = str(len(testsrc_list) + len(testwant_list))
        >>> print(testsrc_list)
        >>> print(testlinenum_list)
        >>> print(func_lineno)
        >>> print(testwant_list)
        >>> print(result)
        4

    Example2:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import ibeis
        >>> func_or_class = ibeis.control.manual_annot_funcs.add_annots
        >>> tup = get_doctest_examples(func_or_class)
        >>> testsrc_list, testwant_list, testlinenum_list, func_lineno, docstr = tup
        >>> result = str(len(testsrc_list) + len(testwant_list))
        >>> print(testsrc_list)
        >>> print(testlinenum_list)
        >>> print(func_lineno)
        >>> print(testwant_list)
        >>> print(result)
        2
    """
    if isinstance(func_or_class, staticmethod):
        func_or_class = func_or_class.__func__
    import utool as ut
    if VERBOSE_TEST:
        print('[util_test][DEPTH 3] get_doctest_examples()')
        print('[util_test] + parsing %r for doctest' % (func_or_class))
        print('[util_test] - name = %r' % (func_or_class.__name__,))
        if hasattr(func_or_class, '__ut_parent_class__'):
            print('[util_test] - __ut_parent_class__ = %r' % (
                func_or_class.__ut_parent_class__,))
    try:
        raise NotImplementedError('FIXME')
        #func_or_class._utinfo['orig_func']
        func_lineno = func_or_class.func_code.co_firstlineno
        # FIXME: doesn't handle decorators well
        #
        # ~~FIXME doesn't account for multiline function definitions
        # actually parse this out~~
        # TODO: rectify with util_insepct get_funcsource with stip def line
        sourcecode = inspect.getsource(func_or_class)
        match = ut.regex_get_match('def [^)]*\\):\n', sourcecode)
        if match is not None:
            num_funcdef_lines = match.group().count('\n')
        else:
            num_funcdef_lines = 1
    except Exception as ex:
        func_lineno = 0
        num_funcdef_lines = 1
        if ut.DEBUG2:
            ut.printex(ex, '[util-test] error getting function line number')

    docstr = ut.get_docstr(func_or_class)

    (testheader_list, testsrc_list, testwant_list,
     testlineoffset_list) = parse_doctest_from_docstr(docstr)
    testlinenum_list = [
        func_lineno + num_funcdef_lines + offset
        for offset in testlineoffset_list
    ]
    if VERBOSE_TEST:
        print('[util_test] L found %d doctests' % (len(testsrc_list),))
    examptup = testsrc_list, testwant_list, testlinenum_list, func_lineno, docstr
    return examptup


def get_module_doctest_tup(testable_list=None, check_flags=True, module=None,
                           allexamples=None, needs_enable=None, N=0,
                           verbose=True, testslow=False):
    """
    Parses module for testable doctesttups
    Depth 2)

    Args:
        testable_list (list): a list of functions (default = None)
        check_flags (bool): (default = True)
        module (None): (default = None)
        allexamples (None): (default = None)
        needs_enable (None): (default = None)
        N (int): (default = 0)
        verbose (bool):  verbosity flag(default = True)
        testslow (bool): (default = False)

    Returns:
        ModuleDoctestTup : (enabled_testtup_list, frame_fpath, all_testflags, module)
            enabled_testtup_list (list): a list of testtup
                testtup (tuple): (name, num, src, want, flag) describes a valid doctest in the module
                    name  (str): test name
                    num   (str): test number of the module / function / class / method
                    src   (str): test source code
                    want  (str): expected test result
                    flag  (str): a valid commandline flag to enable this test
            frame_fpath (str):
                module fpath that will be tested
            module (module):
                the actual module that will be tested
            all_testflags (list):
                the command line arguments that will enable different tests
            exclude_inherited (bool): does not included tests defined in other modules

    CommandLine:
        python -m utool.util_tests --exec-get_module_doctest_tup

    Doctest:
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> #testable_list = [ut.util_import.package_contents]
        >>> testable_list = None
        >>> check_flags = False
        >>> module = ut.util_cplat
        >>> allexamples = False
        >>> needs_enable = None
        >>> N = 0
        >>> verbose = True
        >>> testslow = False
        >>> mod_doctest_tup = get_module_doctest_tup(testable_list, check_flags,
        >>>                                          module, allexamples,
        >>>                                          needs_enable, N, verbose,
        >>>                                          testslow)
        >>> result = ('mod_doctest_tup = %s' % (ut.repr4(mod_doctest_tup, nl=4),))
        >>> print(result)
    """
    #+------------------------
    if VERBOSE_TEST:
        print('[util_test.get_module_doctest tup][DEPTH 2] get_module_doctest tup()')
    import utool as ut  # NOQA
    if needs_enable is None:
        needs_enable = not ut.get_argflag('--enableall')
        #needs_enable = True
    TEST_ALL_EXAMPLES = allexamples or ut.get_argflag(('--allexamples', '--all-examples'))
    parse_testables = True
    if isinstance(testable_list, types.ModuleType):
        # hack
        module = testable_list
        testable_list = []
        testable_name_list = []
    elif testable_list is None:
        testable_list = []
        testable_name_list = []
    else:
        testable_name_list = [ut.get_funcname(func) for func in testable_list]
        parse_testables = False
    #L________________________
    #+------------------------
    # GET_MODULE_DOCTEST_TUP Step 1:
    # Inspect caller module for testable names
    if module is None:
        frame_fpath = '???'
        try:
            # This is a bit finky. Need to be exactly N frames under the main
            # module
            frame = ut.get_parent_frame(N=N)
            main_modname = '__main__'
            frame_name  = frame.f_globals['__name__']
            frame_fpath = frame.f_globals['__file__']
            if frame_name == main_modname:
                module = sys.modules[main_modname]
                entry_modname = ut.get_modname_from_modpath(module.__file__)
                if entry_modname in ['kernprof', 'kernprof-script']:
                    # kernprof clobbers the __main__ variable.
                    # workaround by reimporting the module name
                    import importlib
                    modname = ut.get_modname_from_modpath(frame_fpath)
                    module = importlib.import_module(modname)
        except Exception as ex:
            print(frame.f_globals)
            ut.printex(ex, keys=['frame', 'module'])
            raise
        allexamples = False
    else:
        frame_fpath = module.__file__
        allexamples = True
    #L________________________

    #+------------------------
    # GET_MODULE_DOCTEST_TUP Step 2:
    # --- PARSE TESTABLE FUNCTIONS ---
    # Get testable functions
    if parse_testables:
        try:
            if verbose or VERBOSE_TEST and ut.NOT_QUIET:
                print('[ut.test] Iterating over module funcs')
                print('[ut.test] module =%r' % (module,))

            _testableiter = ut.iter_module_doctestable(module,
                                                       include_inherited=False)
            for key, val in _testableiter:
                if isinstance(val, staticmethod):
                    docstr = inspect.getdoc(val.__func__)
                else:
                    docstr = inspect.getdoc(val)
                docstr = ut.ensure_unicode(docstr)
                if docstr is not None and (docstr.find('Example') >= 0 or docstr.find('Doctest') >= 0):
                    testable_name_list.append(key)
                    testable_list.append(val)
                    if VERBOSE_TEST and ut.NOT_QUIET:
                        print('[ut.test] Testable: %s' % (key,))
                else:
                    if VERBOSE_TEST and ut.NOT_QUIET:
                        if (docstr.find('Example') >= 0 or docstr.find('Doctest') >= 0):
                            print('[ut.test] Ignoring (disabled) : %s' % key)
                        else:
                            print('[ut.test] Ignoring (no Example) : %s' % key)
        except Exception as ex:
            print('FAILED')
            print(docstr)
            ut.printex(ex, keys=['frame'])
            raise

    # OUTPUTS: testable_list
    #L________________________
    #+------------------------
    # GET_MODULE_DOCTEST_TUP Step 3:
    # --- FILTER TESTABLES_---

    # Get testable function examples

    test_sentinals = [
        'ENABLE_DOCTEST',
        'ENABLE_GRID_DOCTEST',
    ]
    if testslow or ut.get_argflag(('--testall', '--testslow', '--test-slow')):
        test_sentinals.append('SLOW_DOCTEST')
    if testslow or ut.get_argflag(('--testall', '--testunstable')):
        test_sentinals.append('UNSTABLE_DOCTEST')

    # FIND THE TEST NAMES REQUESTED
    # Grab sys.argv enabled tests
    cmdline_varargs = ut.get_cmdline_varargs()
    force_enable_testnames_ = cmdline_varargs[:]
    valid_prefix_list = ['--test-', '--exec-', '--dump-']
    #if False:
    for arg in sys.argv:
        for prefix in valid_prefix_list:
            if arg.startswith(prefix):
                testname = arg[len(prefix):]
                #testname = testname.split(':')[0].replace('-', '_')
                force_enable_testnames_.append(testname)
                #break

    # PartA: Fixup names
    # TODO: parse out requested test number here
    # instead of later in the code. See PartB
    force_enable_testnames = []
    for testname in force_enable_testnames_:
        testname = testname.split(':')[0].replace('-', '_')
        testname.split(':')[0].replace('-', '_')
        force_enable_testnames.append(testname)

    def _get_testable_name(testable):
        import utool as ut
        if isinstance(testable, staticmethod):
            testable = testable.__func__
        try:
            testable_name = testable.func_name
        except AttributeError as ex1:
            try:
                testable_name = testable.__name__
            except AttributeError as ex2:
                ut.printex(ex1, ut.repr4(dir(testable)))
                ut.printex(ex2, ut.repr4(dir(testable)))
                raise
        return testable_name

    sorted_testable = sorted(list(set(testable_list)), key=_get_testable_name)
    # Append each testable example
    if VERBOSE_TEST:
        print('Vars:')
        print(' * needs_enable = %r' % (needs_enable,))
        print(' * force_enable_testnames = %r' % (force_enable_testnames,))
        print(' * len(sorted_testable) = %r' % (len(sorted_testable),))
        print(' * cmdline_varargs = %r' % (cmdline_varargs,))
        indenter = ut.Indenter('[FIND_AVAIL]')
        indenter.start()
    # PARSE OUT THE AVAILABLE TESTS FOR EACH REQUEST
    local_testtup_list = []
    for testable in sorted_testable:
        short_testname = _get_testable_name(testable)
        full_testname = None  # Namespaced classname (within module)
        if isinstance(testable, staticmethod):
            testable = testable.__func__
        if hasattr(testable, '__ut_parent_class__'):
            # HACK for getting classname.funcname
            test_namespace = testable.__ut_parent_class__.__name__
            full_testname = test_namespace + '.' + short_testname
        else:
            test_namespace = None
            full_testname = short_testname

        nametup = tuple(ut.unique([full_testname, short_testname]))
        # modpath = ut.get_modpath(module)
        examptup = get_doctest_examples(testable)
        examples, wants, linenums, func_lineno, docstr = examptup
        total_examples = len(examples)
        if total_examples > 0:
            for testno , srcwant_tup in enumerate(zip(examples, wants)):
                src, want = srcwant_tup
                src_ = ut.regex_replace('from __future__ import.*$', '', src)
                test_disabled = not any([src_.find(s) >= 0
                                         for s in test_sentinals])
                skip = (needs_enable and test_disabled and
                        ut.isdisjoint(nametup, force_enable_testnames))
                if not skip:
                    if VERBOSE_TEST:
                        print(' * HACK adding testname=%r to local_testtup_list' % (
                            full_testname,))
                    local_testtup = (nametup, testno, src_, want,
                                     test_namespace, short_testname,
                                     total_examples)
                    local_testtup_list.append(local_testtup)
                else:
                    if VERBOSE_TEST:
                        #print('force_enable_testnames = %r' % (force_enable_testnames,))
                        #print('nametup = %r' % (nametup,))
                        #print('needs_enable = %r' % (needs_enable,))
                        #print('test_disabled = %r' % (test_disabled,))
                        print(' * skipping: %r / %r' % (short_testname,
                                                        full_testname))
        else:
            print('WARNING: no examples in %r for testname=%r' % (frame_fpath,
                                                                  full_testname))
            if verbose:
                print(testable)
                print(examples)
                print(wants)
                print(docstr)
        if VERBOSE_TEST:
            print(' --')
    if VERBOSE_TEST:
        indenter.stop()
    #L________________________
    #+------------------------
    # Get enabled (requested) examples
    if VERBOSE_TEST:
        print('\n-----\n')
        indenter = ut.Indenter('[IS_ENABLED]')
        indenter.start()
        print('Finished parsing available doctests.')
        print('Now we need to find which examples are enabled')
        print('len(local_testtup_list) = %r' % (len(local_testtup_list),))
        print('local_testtup_list.T[0:2].T = %s' %
              ut.repr4(ut.take_column(local_testtup_list, [0, 1])))
        print('sys.argv = %r' % (sys.argv,))
    all_testflags = []
    enabled_testtup_list = []
    distabled_testflags  = []
    subx = ut.get_argval('--subx', type_=int, default=None,
                         help_='Only tests the subxth example')

    def make_valid_testnames(name, num, total):
        return [
            name + ':' + str(num),
            name,
            name + ':' + str(num - total),  # allow negative indices
            # prefix + name.replace('_', '-') + ':' + str(num),
            # prefix + name.replace('_', '-')
        ]

    def make_valid_test_argflags(prefix, name, num, total):
        valid_testnames = make_valid_testnames(name, num, total)
        return [prefix + testname for testname in valid_testnames]

    def check_if_test_requested(nametup, num, total, valid_prefix_list):
        #cmdline_varargs
        if VERBOSE_TEST:
            print('Checking cmdline for %r %r' % (nametup, num))
        valid_argflags = []

        # FIXME: PartB
        # should parse out test number above instead of here
        # See PartA
        mode = None
        veryverb = 0
        # First check positional args
        testflag = None
        for name in nametup:
            valid_testnames = make_valid_test_argflags('', name, num, total)
            if veryverb:
                print('Checking if positional* %r' % (valid_testnames[0:1],))
                print('name = %r' % (name,))
            if any([x in cmdline_varargs for x in valid_testnames]):
                # hack
                mode = 'exec'
                testflag = name
                flag1 = '--exec-' + name + ':' + str(num)
            if testflag is not None:
                if veryverb:
                    print('FOUND POSARG')
                    print(' * testflag = %r' % (testflag,))
                    print(' * num = %r' % (num,))
                break
        # Then check keyword-ish args
        if mode is None:
            for prefix, name in reversed(list(ut.iprod(valid_prefix_list, nametup))):
                valid_argflags = make_valid_test_argflags(prefix, name, num, total)
                if veryverb:
                    print('Checking for flags*: %r' % (valid_argflags[0],))
                flag1 = valid_argflags[0]
                testflag = ut.get_argflag(valid_argflags)
                mode = prefix.replace('-', '')
                if testflag:
                    if veryverb:
                        print("FOUND VARARG")
                    break
            else:
                # print('WARNING NO TEST IS ENABLED %r ' % (nametup,))
                pass
        checktup = flag1, mode, name, testflag
        return checktup

    for local_testtup in local_testtup_list:
        (nametup, num, src, want, shortname, test_namespace, total) = local_testtup
        checktup = check_if_test_requested(nametup, num, total,
                                           valid_prefix_list)
        flag1, mode, name, testflag = checktup
        testenabled = TEST_ALL_EXAMPLES  or not check_flags or testflag
        if subx is not None and subx != num:
            continue
        all_testflags.append(flag1)
        if testenabled:
            if VERBOSE_TEST:
                print('... enabling test')
            testtup = TestTuple(name, num, src, want, flag1,
                                frame_fpath=frame_fpath, mode=mode,
                                total=total, nametup=nametup,
                                shortname=shortname,
                                test_namespace=test_namespace)
            if VERBOSE_TEST:
                print('... ' + str(testtup))
            enabled_testtup_list.append(testtup)
        else:
            if VERBOSE_TEST:
                print('... disabling test')
            distabled_testflags.append(flag1)

    # Attempt to run test without any context
    # This will only work if the function exist and is self contained
    if len(force_enable_testnames_) > 0 and len(enabled_testtup_list) == 0:
        if VERBOSE_TEST:
            print('Forced test did not have a doctest example')
            print('Maybe it can be run without any context')
        import utool as ut
        # assert len(force_enable_testnames) == 1
        test_funcname_ = force_enable_testnames[0]
        if test_funcname_.find('.') != -1:
            test_classname, test_funcname = test_funcname_.split('.')
            class_ = getattr(module, test_classname, None)
            assert class_ is not None
            func_ = getattr(class_, test_funcname, None)
        else:
            test_funcname = test_funcname_
            func_ = getattr(module, test_funcname, None)
        if VERBOSE_TEST:
            print('test_funcname = %r' % (test_funcname,))
            print('func_ = %r' % (func_,))
        if func_ is not None:
            testno = 0
            modname = ut.get_modname_from_modpath(module.__file__)
            want = None
            try:
                if VERBOSE_TEST:
                    print('attempting xdoctest hack')
                # hack to get classmethods to read their example using
                # the xdoctest port
                from xdoctest import docscrape_google
                from xdoctest import core as xdoc_core
                from xdoctest import static_analysis as static
                if func_.__doc__ is None:
                    raise TypeError
                blocks = docscrape_google.split_google_docblocks(func_.__doc__)

                example_blocks = []
                for type_, block in blocks:
                    if type_.startswith('Example') or type_.startswith('Doctest'):
                        example_blocks.append((type_, block))

                if len(example_blocks) == 0:
                    if VERBOSE_TEST:
                        print('xdoctest found no blocks')
                    raise KeyError

                callname = test_funcname_
                hack_testtups = []
                for num, (type_, block) in enumerate(example_blocks):
                    # print('modname = %r' % (modname,))
                    # print('callname = %r' % (callname,))
                    # print('num = %r' % (num,))
                    modpath = static.modname_to_modpath(modname)
                    example = xdoc_core.DocTest(
                        modpath=modpath, callname=callname, docsrc=block, num=num)
                    src = example.format_src(colored=False, want=False,
                                             linenos=False)
                    want = '\n'.join(list(example.wants()))
                    testtup = TestTuple(test_funcname_, num, src, want=want,
                                        flag='--exec-' + test_funcname_,
                                        frame_fpath=frame_fpath, mode='exec',
                                        total=len(example_blocks),
                                        nametup=[test_funcname_])
                    hack_testtups.append(testtup)
                if VERBOSE_TEST:
                    print('hack_testtups = %r' % (hack_testtups,))
                enabled_testtup_list.extend(hack_testtups)
                # src = '\n'.join([line[4:] for line in src.split('\n')])
            except (ImportError, KeyError, TypeError):
                if VERBOSE_TEST:
                    print('xdoctest hack failed')
                # varargs = ut.get_cmdline_varargs()
                varargs = force_enable_testnames[1:]
                # Create dummy doctest
                src = ut.codeblock(
                    '''
                    # DUMMY_DOCTEST
                    from {modname} import *  # NOQA
                    args = {varargs}
                    result = {test_funcname_}(*args)
                    print(result)
                    ''').format(
                        modname=modname,
                        test_funcname_=test_funcname_,
                        varargs=repr(varargs))
                testtup = TestTuple(test_funcname_, testno, src, want=want,
                                    flag='--exec-' + test_funcname_,
                                    frame_fpath=frame_fpath, mode='exec',
                                    total=1, nametup=[test_funcname_])
                enabled_testtup_list.append(testtup)
        else:
            print('function %r was not found in %r' % (test_funcname_, module))

    if VERBOSE_TEST:
        indenter.stop()

    if ut.get_argflag('--list'):
        # HACK: Should probably just return a richer structure
        print('testable_name_list = %s' % (ut.repr4(testable_name_list),))

    mod_doctest_tup = ModuleDoctestTup(enabled_testtup_list, frame_fpath,
                                       all_testflags, module)
    #L________________________
    return mod_doctest_tup


def doctest_was_requested():
    """ lets a  __main__ codeblock know that util_test should do its thing """
    # FIXME; does not handle positinal doctest requests
    valid_prefix_list = ['--exec-', '--test-']
    return '--tf' in sys.argv or any([any([arg.startswith(prefix) for prefix in
                                           valid_prefix_list])
                                      for arg in sys.argv])


def find_doctestable_modnames(dpath_list=None, exclude_doctests_fnames=[],
                              exclude_dirs=[], allow_nonpackages=False):
    """
    Tries to find files with a call to ut.doctest_funcs in the __main__ part
    Implementation is very hacky. Should find a better heuristic

    Args:
        dpath_list (list): list of python package directories
        exclude_doctests_fnames (list): (default = [])
        exclude_dirs (list): (default = [])
        allow_nonpackages (bool): (default = False)

    Returns:
        list: list of filepaths

    CommandLine:
        python -m utool.util_tests find_doctestable_modnames --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA
        >>> import utool as ut
        >>> from os.path import dirname
        >>> dpath_list = [ut.get_module_dir(ut)]
        >>> exclude_doctests_fnames = []
        >>> exclude_dirs = []
        >>> allow_nonpackages = False
        >>> result = find_doctestable_modnames(dpath_list, exclude_doctests_fnames, exclude_dirs, allow_nonpackages)
        >>> print(result)
    """
    import utool as ut
    from os.path import dirname, exists, join
    fpath_list = ut.grep(r'doctest_funcs\(', dpath_list=dpath_list,
                         include_patterns=['*.py'], exclude_dirs=exclude_dirs,
                         recursive=True)[0]
    exclude_doctests_fnames = set(exclude_doctests_fnames)

    def is_not_excluded(fpath):
        return basename(fpath) not in exclude_doctests_fnames

    def is_in_package(fpath):
        return exists(join(dirname(fpath), '__init__.py'))
    fpath_list = list(filter(is_in_package, fpath_list))
    fpath_list = list(filter(is_not_excluded, fpath_list))
    doctest_modname_list = list(map(ut.get_modname_from_modpath, fpath_list))
    doctest_modname_list = ut.unique(doctest_modname_list)
    return doctest_modname_list


def find_untested_modpaths(dpath_list=None, exclude_doctests_fnames=[], exclude_dirs=[]):
    import utool as ut
    fpath_list, lines_list, lxs_list = ut.grep('>>> # ENABLE_DOCTEST',
                                               dpath_list=dpath_list,
                                               include_patterns=['*.py'],
                                               exclude_dirs=exclude_dirs,
                                               recursive=True,
                                               inverse=True)
    exclude_doctests_fnames = set(list(exclude_doctests_fnames) + ['__init__.py'])

    def is_not_excluded(fpath):
        fname = basename(fpath)
        return (not fname.startswith('_')) and fname not in exclude_doctests_fnames
    doctest_modpath_list = list(filter(is_not_excluded, fpath_list))
    #doctest_modname_list = list(map(ut.get_modname_from_modpath, doctest_modpath_list))
    return doctest_modpath_list


def show_was_requested():
    """
    returns True if --show is specified on the commandline or you are in
    IPython (and presumably want some sort of interaction
    """
    try:
        import plottool_ibeis as pt
    except ImportError:
        import plottool as pt
    return pt.show_was_requested()
    #import utool as ut
    #return ut.get_argflag('--show') or ut.inIPython()


try:
    # Use ExitTestException from xdoctest if possible
    from xdoctest import ExitTestException
except ImportError:
    try:
        from xdoctest.core import ExitTestException
    except ImportError:
        class ExitTestException(Exception):
            pass


def qt4ensure():
    try:
        import plottool_ibeis as pt
    except ImportError:
        import plottool as pt
    pt.qtensure()


def qtensure():
    try:
        import plottool_ibeis as pt
    except ImportError:
        import plottool as pt
    pt.qtensure()


def quit_if_noshow():
    import utool as ut
    saverequest = ut.get_argval('--save', default=None)
    if not (saverequest or ut.get_argflag(('--show', '--save')) or ut.inIPython()):
        raise ExitTestException('This should be caught gracefully by ut.run_test')


def show_if_requested():
    try:
        import plottool_ibeis as pt
    except ImportError:
        import plottool as pt
    pt.show_if_requested(N=2)


def find_testfunc(module, test_funcname, ignore_prefix=[], ignore_suffix=[],
                  func_to_module_dict={}, return_mod=False):
    import utool as ut
    if isinstance(module, six.string_types):
        module = ut.import_modname(module)
    modname_list = ut.package_contents(module, ignore_prefix=ignore_prefix,
                                       ignore_suffix=ignore_suffix)
    # Get only the modules already imported
    have_modnames = [modname_ for modname_ in modname_list
                     if modname_ in sys.modules]
    #missing_modnames = [modname for modname in modname_list
    #                    if modname not in sys.modules]
    module_list = ut.dict_take(sys.modules, have_modnames)
    # Search for the module containing the function
    test_func = None
    test_module = None
    test_classname = None
    if test_funcname.find('.') != -1:
        test_classname, test_funcname = test_funcname.split('.')
    if test_funcname.find(':') != -1:
        test_funcname, testno = test_funcname.split(':')
        testno = int(testno)
    else:
        testno = 0
    if test_classname is None:
        for module_ in module_list:
            #test_funcname = 'find_installed_tomcat'
            if test_funcname in module_.__dict__:
                test_module = module_
                test_func = test_module.__dict__[test_funcname]
                break
    else:
        for module_ in module_list:
            #test_funcname = 'find_installed_tomcat'
            if test_classname in module_.__dict__:
                test_module = module_
                test_class = test_module.__dict__[test_classname]
                test_func = getattr(test_class, test_funcname)
                # test_class.__dict__[test_funcname]

    if test_func is None:
        print('Did not find any function named %r ' % (test_funcname,))
        print('Searched ' + ut.repr4([mod.__name__ for mod in module_list]))
    if return_mod:
        return test_func, testno, test_module
    else:
        return test_func, testno


def get_module_completions(module):
    import utool as ut
    test_tuples = ut.get_package_testables(module)
    testnames = ut.instancelist(test_tuples).name
    return testnames


#def autocomplete_hook(module):
#    """
#    # https://argcomplete.readthedocs.io/en/latest/#activating-global-completion%20argcomplete
#    pip install argcomplete

#    Need to put
#    PYTHON_ARGCOMPLETE_OK at begining of file

#    pip install argcomplete
#    activate-global-python-argcomplete

#    eval "$(register-python-argcomplete your_script)"
#    register-python-argcomplete ibeis
#    eval "$(register-python-argcomplete ibeis)"
#    """
#    #if len(sys.argv) < 3:
#    #    sys.exit(1)
#    try:
#        # hook
#        import argparse as ap
#        import argcomplete
#    except ImportError:
#        pass
#    else:
#        parser = ap.ArgumentParser()
#        testnames = get_module_completions(module)
#        testnames = ['foo', 'foo2']
#        parser.add_argument('position1', choices=[testnames])
#        argcomplete.autocomplete(parser)
#        args = parser.parse_args()


def main_function_tester(module, ignore_prefix=[], ignore_suffix=[],
                         test_funcname=None, func_to_module_dict={}):
    """
    Allows a shorthand for __main__ packages of modules to run tests with
    unique function names
    """
    import utool as ut
    ut.colorprint('[utool] main_function_tester', 'yellow')

    if ut.get_argflag('--list-testfuncs'):
        print('Listing testfuncs')
        test_tuples = ut.get_package_testables(module)
        result = ut.repr3(test_tuples)
        print(result)

    #autocomplete_hook(module)

    if ut.get_argflag('--update-bashcomplete'):
        # http://stackoverflow.com/questions/427472/line-completion-with-custom-commands
        print('Listing testfuncs')
        testnames = get_module_completions(module)
        modname = module if isinstance(module, six.string_types) else module.__name__
        line = 'complete -W "%s" "%s"' % (' '.join(testnames), modname)
        bash_completer = ut.unixjoin(ut.ensure_app_resource_dir('ibeis'), 'ibeis_bash_complete.sh')
        ut.writeto(bash_completer, line)
        print('ADD TO BASHRC\nsource %s' % (bash_completer,))
        #print(line)
        sys.exit(ut.EXIT_SUCCESS)

    if ut.get_argflag('--make-bashcomplete'):
        # http://stackoverflow.com/questions/427472/line-completion-with-custom-commands
        print('Listing testfuncs')
        testnames = get_module_completions(module)
        modname = module if isinstance(module, six.string_types) else module.__name__
        line = 'complete -W "%s" "%s"' % (' '.join(testnames), modname)
        print('add the following line to your bashrc')
        print(line)
        sys.exit(ut.EXIT_SUCCESS)

    test_funcname = ut.get_argval(
        ('--test-func', '--tfunc', '--tf', '--testfunc'),
        type_=str, default=test_funcname,
        help_='specify a function to doctest')
    if test_funcname is None:
        cmdline_varags = ut.get_cmdline_varargs()
        if VERBOSE_TEST:
            print('Checking varargs')
            print('cmdline_varags = %r' % (cmdline_varags,))
        if len(cmdline_varags) > 0:
            test_funcname = cmdline_varags[0]
    print('test_funcname = %r' % (test_funcname,))

    if test_funcname in func_to_module_dict:
        modname = func_to_module_dict[test_funcname]
        ut.import_modname(modname)

    if test_funcname is not None:
        #locals_ = {}
        ut.inject_colored_exceptions()
        # print('[utool] __main__ Begin Function Test')
        print('[utool] __main__ Begin Function Test')
        test_func, testno, test_mod = find_testfunc(
            module, test_funcname, ignore_prefix, ignore_suffix,
            func_to_module_dict, return_mod=True)

        if test_func is not None:
            globals_ = {}
            try:
                func_globals = ut.get_funcglobals(test_func)
                globals_.update(func_globals)
            except AttributeError:
                pass
            tests = ut.get_doctest_examples(test_func)[0]
            if len(tests) > 0:
                testsrc = tests[testno]
            else:
                # Create dummy doctest
                modname = ut.get_modname_from_modpath(ut.get_modpath(test_mod))
                varargs = cmdline_varags[1:]
                testsrc = ut.codeblock(
                    '''
                    # DUMMY_DOCTEST
                    from {modname} import *  # NOQA
                    args = {varargs}
                    result = {test_funcname_}(*args)
                    print(result)
                    ''').format(
                        modname=modname,
                        test_funcname_=test_funcname,
                        varargs=repr(varargs))
                # testtup = TestTuple(test_funcname_, testno, src, want=want,
                #                     flag='--exec-' + test_funcname_,
                #                     frame_fpath=frame_fpath, mode='exec',
                #                     total=1, nametup=[test_funcname_])

            if ut.get_argflag(('--cmd', '--embed')):
                # TODO RECTIFY WITH EXEC DOCTEST
                testsrc = _cmd_modify_src(testsrc)
            # _exec_doctest(doctest_src)
            # doctest_src = ut.indent(testsrc, '>>> ')
            # Add line numbers
            doctest_src = ut.number_text_lines(testsrc)
            colored_src = ut.highlight_code(doctest_src)
            print('testsrc = \n%s' % (colored_src,))
            try:
                code = compile(testsrc, '<string>', 'exec')
                exec(code, globals_)  # , locals_)
            except ExitTestException:
                print('Test exited before show')
                pass
            retcode = ut.EXIT_SUCCESS
            print('Finished function test.')
        else:
            retcode = ut.EXIT_FAILURE
            print('Did not find any function named %r ' % (test_funcname,))
        if ut.util_inject.PROFILING:
            ut.dump_profile_text()
        print('...exiting')
        sys.exit(retcode)


def _cmd_modify_src(testsrc):
    # testsrc = 'import IPython; import utool as ut; ut.qtensure()\n' + testsrc
    # testsrc += '\nimport IPython; IPython.embed()'
    # testsrc += '\nimport utool as ut; ut.embed()'
    import utool as ut
    pline = 'print(%r)' % ut.highlight_code(ut.indent(testsrc, '>>> '))
    testsrc += '\nimport utool as ut; ' + pline + '; ut.qtensure(); ut.embed()'
    # testsrc += '\nimport utool as ut; ut.qtensure(); ut.embed()'
    return testsrc


def execute_doctest(func, testnum=0, module=None):
    """
    Execute a function doctest. Can optionaly specify func name and module to
    run from ipython notebooks.

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_tests import *  # NOQA

    IPython:
        import utool as ut
        ut.execute_doctest(func='dummy_example_depcacahe', module='dtool.example_depcache')
    """
    import utool as ut
    if isinstance(func, six.string_types):
        funcname = func
        if isinstance(module, six.string_types):
            modname = module
            module = ut.import_modname(modname)
        func, _testno = find_testfunc(module, funcname, ignore_prefix=[],
                                      ignore_suffix=[])
    # TODO RECTIFY WITH EXEC DOCTEST
    globals_ = {}
    testsrc = ut.get_doctest_examples(func)[0][testnum]
    # colored_src = ut.highlight_code(ut.indent(testsrc, '>>> '))
    doctest_src = ut.indent(testsrc, '>>> ')
    doctest_src = '\n'.join([
        '%3d %s' % (count, line)
        for count, line in enumerate(doctest_src.splitlines(), start=1)])
    colored_src = ut.highlight_code(doctest_src)
    print('testsrc = \n%s' % (colored_src,))
    try:
        code = compile(testsrc, '<string>', 'exec')
        exec(code, globals_)
    except ExitTestException:
        print('Test exited before show')


if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_tests; utool.doctest_funcs(utool.util_tests)"
        python -m utool.util_tests
        python -m utool.util_tests --allexamples
        python -m utool.util_tests

        python -c "import utool; utool.doctest_funcs(module=utool.util_tests, needs_enable=False)"
        /model/preproc/preproc_chip.py --allexamples
    """
    import multiprocessing
    import utool as ut  # NOQA
    multiprocessing.freeze_support()
    #doctest_funcs()
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import builtins
import inspect
import textwrap
import six
import sys
import functools
import os
from utool import util_print
from utool import util_time
from utool import util_iter
from utool import util_dbg
from utool import util_arg
from utool import util_type
from utool import util_inject
from utool._internal import meta_util_six
(print, rrr, profile) = util_inject.inject2(__name__, '[decor]')

if util_type.HAVE_NUMPY:
    import numpy as np

# Commandline to toggle certain convinience decorators
SIG_PRESERVE = util_arg.get_argflag('--sigpreserve')
#SIG_PRESERVE = not util_arg.SAFE or util_arg.get_argflag('--sigpreserve')
ONEX_REPORT_INPUT = '--onex-report-input' in sys.argv
#IGNORE_TRACEBACK = '--smalltb' in sys.argv or '--ignoretb' in sys.argv
# FIXME: dupliated in _internal/py2_syntax_funcs
IGNORE_TRACEBACK = not ('--nosmalltb' in sys.argv or '--noignoretb' in sys.argv)
#if util_arg.STRICT:
#    IGNORE_TRACEBACK = False

# do not ignore traceback when profiling
PROFILING = hasattr(builtins, 'profile')
UNIQUE_NUMPY = True
NOINDENT_DECOR = False

#os.environ.get('UTOOL_AUTOGEN_SPHINX_RUNNING', 'OFF')

#def composed(*decs):
#    """ combines multiple decorators """
#    def deco(f):
#        for dec in reversed(decs):
#            f = dec(f)
#        return f
#    return deco


def test_ignore_exec_traceback():
    r"""
    CommandLine:
        python -m utool.util_decor --test-test_ignore_exec_traceback

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_decor import *  # NOQA
        >>> result = test_ignore_exec_traceback()
        >>> print(result)
    """
    import utool as ut
    @ut.indent_func
    def foobar():
        print('foobar')
        raise AssertionError('This error is exepcted')

    try:
        print('printing foobar')
        foobar()
    except AssertionError as ex:
        #import sys
        #exc_type, exc_value, exc_traceback = sys.exc_info()
        #print(exc_traceback)
        # TODO: ensure decorators are not printed in stack trace
        ut.printex(ex, 'There is no error. This is a test', tb=True)


if six.PY2:
    # Use version that has special python2 only syntax.
    # can not include it here for that reason
    from utool._internal import py2_syntax_funcs
    ignores_exc_tb = py2_syntax_funcs.ignores_exc_tb
else:
    def ignores_exc_tb(*args, **kwargs):
        """
        PYTHON 3 VERSION

        ignore_exc_tb decorates a function and remove both itself
        and the function from any exception traceback that occurs.

        This is useful to decorate other trivial decorators
        which are polluting your stacktrace.

        if IGNORE_TRACEBACK is False then this decorator does nothing
        (and it should do nothing in production code!)

        References:
            https://github.com/jcrocholl/pep8/issues/34  # NOQA
            http://legacy.python.org/dev/peps/pep-3109/
        """
        outer_wrapper = kwargs.get('outer_wrapper', True)
        def ignores_exc_tb_closure(func):
            # HACK JUST TURN THIS OFF
            return func

            if not IGNORE_TRACEBACK:
                # if the global enforces that we should not ignore anytracebacks
                # then just return the original function without any modifcation
                return func
            #@wraps(func)
            def wrp_noexectb(*args, **kwargs):
                try:
                    #import utool
                    #if utool.DEBUG:
                    #    print('[IN IGNORETB] args=%r' % (args,))
                    #    print('[IN IGNORETB] kwargs=%r' % (kwargs,))
                    return func(*args, **kwargs)
                except Exception:
                    # PYTHON 3.3 NEW METHODS
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    # Code to remove this decorator from traceback
                    # Remove two levels to remove this one as well
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    try:
                        exc_traceback = exc_traceback.tb_next
                        exc_traceback = exc_traceback.tb_next
                    except Exception:
                        pass
                    ex = exc_type(exc_value)
                    ex.__traceback__ = exc_traceback
                    raise ex
            if outer_wrapper:
                wrp_noexectb = preserve_sig(wrp_noexectb, func)
            return wrp_noexectb
        if len(args) == 1:
            # called with one arg means its a function call
            func = args[0]
            return ignores_exc_tb_closure(func)
        else:
            # called with no args means kwargs as specified
            return ignores_exc_tb_closure


# NEW PYTHON 2.7/3 VERSION
#def ignores_exc_tb(*args, **kwargs):
#    """
#    ignore_exc_tb decorates a function and remove both itself
#    and the function from any exception traceback that occurs.

#    This is useful to decorate other trivial decorators
#    which are polluting your stacktrace.

#    if IGNORE_TRACEBACK is False then this decorator does nothing
#    (and it should do nothing in production code!)

#    References:
#        https://github.com/jcrocholl/pep8/issues/34  # NOQA
#        http://legacy.python.org/dev/peps/pep-3109/
#    """
#    outer_wrapper = kwargs.get('outer_wrapper', True)
#    def ignores_exc_tb_closure(func):
#        if not IGNORE_TRACEBACK:
#            # if the global enforces that we should not ignore anytracebacks
#            # then just return the original function without any modifcation
#            return func
#        if six.PY2:
#            #python2_func = """
#            common_wrp_noexcept_tb = """
#            def wrp_noexectb(*args, **kwargs):
#                try:
#                    return func(*args, **kwargs)
#                except Exception:
#                    exc_type, exc_value, exc_traceback = sys.exc_info()
#                    # Code to remove this decorator from traceback
#                    # Remove two levels to remove this one as well
#                    exc_type, exc_value, exc_traceback = sys.exc_info()
#                    try:
#                        exc_traceback = exc_traceback.tb_next
#                        exc_traceback = exc_traceback.tb_next
#                    except Exception:
#                        pass
#            """
#            if six.PY2:
#                python2_reraise = """
#                    raise exc_type, exc_value, exc_traceback
#                """
#                six_reraise = python2_reraise
#            elif six.PY3:
#                python3_reraise = """
#                    ex = exc_type(exc_value)
#                    ex.__traceback__ = exc_traceback
#                    raise ex
#                """
#                six_reraise = python3_reraise
#            wrp_noexcept_tb_codeblock = common_wrp_noexcept_tb + six_reraise
#            globals_ = globals()
#            locals_ = locals()
#            six.exec_(wrp_noexcept_tb_codeblock, globals_, locals_)
#            wrp_noexectb = locals_['wrp_noexectb']
#        if outer_wrapper:
#            wrp_noexectb = preserve_sig(wrp_noexectb, func)
#        return wrp_noexectb
#    if len(args) == 1:
#        # called with one arg means its a function call
#        func = args[0]
#        return ignores_exc_tb_closure(func)
#    else:
#        # called with no args means kwargs as specified
#        return ignores_exc_tb_closure


def on_exception_report_input(func_=None, force=False, keys=None):
    """
    If an error is thrown in the scope of this function's stack frame then the
    decorated function name and the arguments passed to it will be printed to
    the utool print function.
    """
    def _closure_onexceptreport(func):
        if not ONEX_REPORT_INPUT and not force:
            return func
        @ignores_exc_tb(outer_wrapper=False)
        #@wraps(func)
        def wrp_onexceptreport(*args, **kwargs):
            try:
                #import utool
                #if utool.DEBUG:
                #    print('[IN EXCPRPT] args=%r' % (args,))
                #    print('[IN EXCPRPT] kwargs=%r' % (kwargs,))
                return func(*args, **kwargs)
            except Exception as ex:
                from utool import util_str
                print('ERROR occured! Reporting input to function')
                if keys is not None:
                    from utool import util_inspect
                    from utool import util_list
                    from utool import util_dict
                    argspec = util_inspect.get_func_argspec(func)
                    in_kwargs_flags = [key in kwargs for key in keys]
                    kwarg_keys = util_list.compress(keys, in_kwargs_flags)
                    kwarg_vals = [kwargs.get(key) for key in kwarg_keys]
                    flags = util_list.not_list(in_kwargs_flags)
                    arg_keys = util_list.compress(keys, flags)
                    arg_idxs = [argspec.args.index(key) for key in arg_keys]
                    num_nodefault = len(argspec.args) - len(argspec.defaults)
                    default_vals = (([None] * (num_nodefault)) +
                                    list(argspec.defaults))
                    args_ = list(args) + default_vals[len(args) + 1:]
                    arg_vals = util_list.take(args_, arg_idxs)
                    requested_dict = dict(util_list.flatten(
                        [zip(kwarg_keys, kwarg_vals), zip(arg_keys, arg_vals)]))
                    print('input dict = ' + util_str.repr4(
                        util_dict.dict_subset(requested_dict, keys)))
                    # (print out specific keys only)
                    pass
                arg_strs = ', '.join([repr(util_str.truncate_str(str(arg)))
                                      for arg in args])
                kwarg_strs = ', '.join([
                    util_str.truncate_str('%s=%r' % (key, val))
                    for key, val in six.iteritems(kwargs)])
                msg = ('\nERROR: funcname=%r,\n * args=%s,\n * kwargs=%r\n' % (
                    meta_util_six.get_funcname(func), arg_strs, kwarg_strs))
                msg += ' * len(args) = %r\n' % len(args)
                msg += ' * len(kwargs) = %r\n' % len(kwargs)
                util_dbg.printex(ex, msg, pad_stdout=True)
                raise
        wrp_onexceptreport = preserve_sig(wrp_onexceptreport, func)
        return wrp_onexceptreport
    if func_ is None:
        return _closure_onexceptreport
    else:
        return _closure_onexceptreport(func_)


def debug_function_exceptions(func):
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            import utool as ut
            ut.printex(ex)
            import inspect  # NOQA
            trace = inspect.trace()
            locals_ = trace[-1][0].f_locals
            print('-- <TRACE LOCALS> --')
            for level, t in enumerate(trace[1:]):
                frame = t[0]
                locals_ = frame.f_locals
                local_repr_dict = {key: ut.trunc_repr(val)
                                   for key, val in locals_.items()}
                print('LOCALS LEVEL %d' % (level,))
                print(ut.repr3(local_repr_dict, strvals=True, nl=1))
            print('-- </TRACE LOCALS> --')
            #import utool
            #utool.embed()
            raise
    return _wrapper


#class DebugContext(object):
#    def __enter__():
#        pass
#    def __exit__(self, exc_type, exc_value, exc_traceback):
#        pass


def _indent_decor(lbl):
    """
    does the actual work of indent_func
    """
    def closure_indent(func):
        if util_arg.TRACE:
            @ignores_exc_tb(outer_wrapper=False)
            #@wraps(func)
            def wrp_indent(*args, **kwargs):
                with util_print.Indenter(lbl):
                    print('    ...trace[in]')
                    ret = func(*args, **kwargs)
                    print('    ...trace[out]')
                    return ret
        else:
            @ignores_exc_tb(outer_wrapper=False)
            #@wraps(func)
            def wrp_indent(*args, **kwargs):
                with util_print.Indenter(lbl):
                    ret = func(*args, **kwargs)
                    return ret
        wrp_indent_ = ignores_exc_tb(wrp_indent)
        wrp_indent_ = preserve_sig(wrp_indent, func)
        return wrp_indent_
    return closure_indent


def indent_func(input_):
    """
    Takes either no arguments or an alias label
    """
    if isinstance(input_, six.string_types):
        # A label was specified
        lbl = input_
        return _indent_decor(lbl)
    elif isinstance(input_, (bool, tuple)):
        # Allow individually turning of of this decorator
        func = input_
        return func
    else:
        # Use the function name as the label
        func = input_
        lbl = '[' + meta_util_six.get_funcname(func) + ']'
        return _indent_decor(lbl)(func)


def tracefunc_xml(func):
    """
    Causes output of function to be printed in an XML style block
    """
    funcname = meta_util_six.get_funcname(func)
    def wrp_tracefunc2(*args, **kwargs):
        verbose = kwargs.get('verbose', True)
        if verbose:
            print('<%s>' % (funcname,))
        with util_print.Indenter('    '):
            ret = func(*args, **kwargs)
        if verbose:
            print('</%s>' % (funcname,))
        return ret
    wrp_tracefunc2_ = ignores_exc_tb(wrp_tracefunc2)
    wrp_tracefunc2_ = preserve_sig(wrp_tracefunc2_, func)
    return wrp_tracefunc2_

#----------


def accepts_scalar_input(func):
    """
    DEPRICATE in favor of accepts_scalar_input2
    only accepts one input as vector

    accepts_scalar_input is a decorator which expects to be used on class
    methods.  It lets the user pass either a vector or a scalar to a function,
    as long as the function treats everything like a vector. Input and output
    is sanitized to the user expected format on return.

    Args:
        func (func):

    Returns:
        func: wrp_asi

    CommandLine:
        python -m utool.util_decor --test-accepts_scalar_input

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_decor import *  # NOQA
        >>> @accepts_scalar_input
        ... def foobar(self, list_):
        ...     return [x + 1 for x in list_]
        >>> self = None  # dummy self because this decorator is for classes
        >>> assert 2 == foobar(self, 1)
        >>> assert [2, 3] == foobar(self, [1, 2])
    """
    #@on_exception_report_input
    @ignores_exc_tb(outer_wrapper=False)
    #@wraps(func)
    def wrp_asi(self, input_, *args, **kwargs):
        #if HAVE_PANDAS:
        #    if isinstance(input_, (pd.DataFrame, pd.Series)):
        #        input_ = input_.values
        if util_iter.isiterable(input_):
            # If input is already iterable do default behavior
            return func(self, input_, *args, **kwargs)
        else:
            # If input is scalar, wrap input, execute, and unpack result
            #ret = func(self, (input_,), *args, **kwargs)
            ret = func(self, [input_], *args, **kwargs)
            if ret is not None:
                return ret[0]
    wrp_asi = preserve_sig(wrp_asi, func)
    return wrp_asi


def accepts_scalar_input2(argx_list=[0], outer_wrapper=True):
    r"""
    FIXME: change to better name. Complete implementation.

    used in IBEIS setters

    accepts_scalar_input2 is a decorator which expects to be used on class
    methods.  It lets the user pass either a vector or a scalar to a function,
    as long as the function treats everything like a vector. Input and output
    is sanitized to the user expected format on return.

    Args:
        argx_list (list): indexes of args that could be passed in as scalars to
            code that operates on lists. Ensures that decorated function gets
            the argument as an iterable.
    """
    assert isinstance(argx_list, (list, tuple)), (
        'accepts_scalar_input2 must be called with argument positions')

    def closure_asi2(func):
        #@on_exception_report_input
        @ignores_exc_tb(outer_wrapper=False)
        def wrp_asi2(self, *args, **kwargs):
            # Hack in case wrapping a function with varargs
            argx_list_ = [argx for argx in argx_list if argx < len(args)]
            __assert_param_consistency(args, argx_list_)
            if all([util_iter.isiterable(args[ix]) for ix in argx_list_]):
                # If input is already iterable do default behavior
                return func(self, *args, **kwargs)
            else:
                # If input is scalar, wrap input, execute, and unpack result
                args_wrapped = [(arg,) if ix in argx_list_ else arg
                                for ix, arg in enumerate(args)]
                ret = func(self, *args_wrapped, **kwargs)
                if ret is not None:
                    return ret[0]
        if outer_wrapper:
            wrp_asi2 = on_exception_report_input(preserve_sig(wrp_asi2, func))
        return wrp_asi2
    return closure_asi2


def __assert_param_consistency(args, argx_list_):
    """
    debugging function for accepts_scalar_input2
    checks to make sure all the iterable inputs are of the same length
    """
    if util_arg.NO_ASSERTS:
        return
    if len(argx_list_) == 0:
        return True
    argx_flags = [util_iter.isiterable(args[argx]) for argx in argx_list_]
    try:
        assert all([argx_flags[0] == flag for flag in argx_flags]), (
            'invalid mixing of iterable and scalar inputs')
    except AssertionError as ex:
        print('!!! ASSERTION ERROR IN UTIL_DECOR !!!')
        for argx in argx_list_:
            print('[util_decor] args[%d] = %r' % (argx, args[argx]))
        raise ex


def accepts_scalar_input_vector_output(func):
    """
    DEPRICATE IN FAVOR OF accepts_scalar_input2

    accepts_scalar_input_vector_output

    Notes:
        Input:                           Excpeted Output 1to1   Expected Output 1toM
            scalar         : 1           x                      [X]
            n element list : [1, 2, 3]   [x, y, z]              [[X], [Y], [Z]]
            1 element list : [1]         [x]                    [[X]]
            0 element list : []          []                     []
        There seems to be no real issue here, I be the thing that tripped me up
        was when using sql and getting multiple columns that returned the
        values inside of the N-tuple whereas when you get one column you get
        one element inside of a 1-tuple, no that still makes sense.  There was
        something where when you couln't unpack it becuase it was already
        empty...
    """
    @ignores_exc_tb(outer_wrapper=False)
    #@wraps(func)
    def wrp_asivo(self, input_, *args, **kwargs):
        #import utool
        #if utool.DEBUG:
        #    print('[IN SIVO] args=%r' % (args,))
        #    print('[IN SIVO] kwargs=%r' % (kwargs,))
        if util_iter.isiterable(input_):
            # If input is already iterable do default behavior
            return func(self, input_, *args, **kwargs)
        else:
            # If input is scalar, wrap input, execute, and unpack result
            result = func(self, (input_,), *args, **kwargs)
            # The output length could be 0 on a scalar input
            if len(result) == 0:
                return []
            else:
                assert len(result) == 1, 'error in asivo'
                return result[0]
    return wrp_asivo

# TODO: Rename to listget_1to1 1toM etc...
getter_1to1 = accepts_scalar_input
getter_1toM = accepts_scalar_input_vector_output
#----------


def accepts_numpy(func):
    """ Allows the first input to be a numpy array and get result in numpy form """
    #@ignores_exc_tb
    #@wraps(func)
    def wrp_accepts_numpy(self, input_, *args, **kwargs):
        if not (util_type.HAVE_NUMPY and isinstance(input_, np.ndarray)):
            # If the input is not numpy, just call the function
            return func(self, input_, *args, **kwargs)
        else:
            # TODO: use a variant of util_list.unflat_unique_rowid_map
            # If the input is a numpy array, and return the output with the same
            # shape as the input
            if UNIQUE_NUMPY:
                # Remove redundant input (because we are passing it to SQL)
                input_list, inverse_unique = np.unique(input_, return_inverse=True)
            else:
                input_list = input_.flatten()
            # Call the function in list format
            # TODO: is this necessary?
            input_list = input_list.tolist()
            output_list = func(self, input_list, *args, **kwargs)
            # Put the output back into numpy
            if UNIQUE_NUMPY:
                # Reconstruct redundant queries
                output_arr = np.array(output_list)[inverse_unique]
                output_shape = tuple(list(input_.shape) + list(output_arr.shape[1:]))
                return np.array(output_arr).reshape(output_shape)
            else:
                return np.array(output_list).reshape(input_.shape)
    wrp_accepts_numpy = preserve_sig(wrp_accepts_numpy, func)
    return wrp_accepts_numpy


def memoize_nonzero(func):
    """
    Memoization decorator for functions taking a nonzero number of arguments.

    References:
        http://code.activestate.com/recipes/578231-fastest-memoization-decorator
    """
    class _memorizer(dict):
        def __init__(self, func):
            self.func = func
        def __call__(self, *args):
            return self[args]
        def __missing__(self, key):
            ret = self[key] = self.func(*key)
            return ret
    return _memorizer(func)


def memoize_single(func):
    """ Memoization decorator for a function taking a single argument

    References:
        http://code.activestate.com/recipes/578231-fastest-memoization-decorator
    """
    class memodict_single(dict):
        def __missing__(self, key):
            ret = self[key] = func(key)
            return ret
    return memodict_single().__getitem__


def memoize_zero(func):
    """ Memoization decorator for a function taking no arguments """
    wrp_memoize_single = memoize_single(func)
    def wrp_memoize_zero():
        return wrp_memoize_single(None)
    return wrp_memoize_zero


def memoize(func):
    """
    simple memoization decorator

    References:
        https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize

    Args:
        func (function):  live python function

    Returns:
        func:

    CommandLine:
        python -m utool.util_decor memoize

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_decor import *  # NOQA
        >>> import utool as ut
        >>> closure = {'a': 'b', 'c': 'd'}
        >>> incr = [0]
        >>> def foo(key):
        >>>     value = closure[key]
        >>>     incr[0] += 1
        >>>     return value
        >>> foo_memo = memoize(foo)
        >>> assert foo('a') == 'b' and foo('c') == 'd'
        >>> assert incr[0] == 2
        >>> print('Call memoized version')
        >>> assert foo_memo('a') == 'b' and foo_memo('c') == 'd'
        >>> assert incr[0] == 4
        >>> assert foo_memo('a') == 'b' and foo_memo('c') == 'd'
        >>> print('Counter should no longer increase')
        >>> assert incr[0] == 4
        >>> print('Closure changes result without memoization')
        >>> closure = {'a': 0, 'c': 1}
        >>> assert foo('a') == 0 and foo('c') == 1
        >>> assert incr[0] == 6
        >>> assert foo_memo('a') == 'b' and foo_memo('c') == 'd'
    """
    cache = func._util_decor_memoize_cache = {}
    # @functools.wraps(func)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    memoizer = preserve_sig(memoizer, func)
    memoizer.cache = cache
    return memoizer


def interested(func):
    @indent_func
    #@ignores_exc_tb
    #@wraps(func)
    def wrp_interested(*args, **kwargs):
        sys.stdout.write('#\n')
        sys.stdout.write('#\n')
        sys.stdout.write(
            '<!INTERESTED>: ' + meta_util_six.get_funcname(func) + '\n')
        print('INTERESTING... ' + (' ' * 30) + ' <----')
        return func(*args, **kwargs)
    return wrp_interested


def tracefunc(func):
    lbl = '[trace.' + meta_util_six.get_funcname(func) + ']'
    def wrp_tracefunc(*args, **kwargs):
        print(lbl + ' +--- ENTER ---')
        with util_print.Indenter(lbl + ' |'):
            ret = func(*args, **kwargs)
        print(lbl + ' L___ EXIT ____')
        return ret
    return wrp_tracefunc


def show_return_value(func):
    from utool.util_str import func_str
    #@wraps(func)
    def wrp_show_return_value(*args, **kwargs):
        ret = func(*args, **kwargs)
        #print('%s(*%r, **%r) returns %r' % (meta_util_six.get_funcname(func), args, kwargs, rv))
        print(func_str(func, args, kwargs)  + ' -> ret=%r' % (ret,))
        return ret
    return wrp_show_return_value


def time_func(func):
    #@wraps(func)
    def wrp_time(*args, **kwargs):
        with util_time.Timer(meta_util_six.get_funcname(func)):
            return func(*args, **kwargs)
    wrp_time = preserve_sig(wrp_time, func)
    return wrp_time


#def rename_func(newname):
#    import utool as ut
#    return ut.partial(ut.set_funcname, newname=newname)


#class copy_argspec(object):
#    """
#    copy_argspec is a signature modifying decorator.
#    Specifically, it copies the signature from `source_func` to the wrapper, and
#    the wrapper will call the original function (which should be using *args,
#    **kwds).  The argspec, docstring, and default values are copied from
#    src_func, and __module__ and __dict__ from tgt_func.
#    .. References
#    http://stackoverflow.com/questions/18625510/how-can-i-programmatically-change-the-argspec-of-a-function-not-in-a-python-de
#    """
#    def __init__(self, src_func):
#        self.argspec = inspect.getargspec(src_func)
#        self.src_doc = src_func.__doc__
#        self.src_defaults = src_func.func_defaults
#    def __call__(self, tgt_func):
#        try:
#            tgt_argspec = inspect.getargspec(tgt_func)
#            need_self = False
#            if len(tgt_argspec) > 0 and len(tgt_argspec[0]) > 0 and tgt_argspec[0][0] == 'self':
#                need_self = True
#            name = tgt_func.__name__
#            argspec = self.argspec
#            if len(argspec) > 0 and len(argspec[0]) > 0 and argspec[0][0] == 'self':
#                need_self = False
#            if need_self:
#                newargspec = (['self'] + argspec[0],) + argspec[1:]
#            else:
#                newargspec = argspec
#            signature = inspect.formatargspec(formatvalue=lambda val: "",
#                                              *newargspec)[1:-1]
#            new_func = (
#                'def _wrapper_({signature}):\n'
#                '    return {tgt_func}({signature})'
#            ).format(signature=signature, tgt_func='tgt_func')
#            evaldict = {'tgt_func' : tgt_func}
#            exec new_func in evaldict
#            wrapped = evaldict['_wrapper_']
#            wrapped.__name__ = name
#            wrapped.__doc__ = self.src_doc
#            wrapped.func_defaults = self.src_defaults
#            wrapped.__module__ = tgt_func.__module__
#            wrapped.__dict__ = tgt_func.__dict__
#            return wrapped
#        except Exception as ex:
#            util_dbg.printex(ex, 'error wrapping: %r' % (tgt_func,))
#            raise


def lazyfunc(func):
    """
    Returns a memcached version of a function
    """
    closuremem_ = [{}]
    def wrapper(*args, **kwargs):
        mem = closuremem_[0]
        key = (repr(args), repr(kwargs))
        try:
            return mem[key]
        except KeyError:
            mem[key] = func(*args, **kwargs)
        return mem[key]
    return wrapper


def apply_docstr(docstr_func):
    """
    Changes docstr of one functio to that of another
    """
    def docstr_applier(func):
        #docstr = meta_util_six.get_funcdoc(docstr_func)
        #meta_util_six.set_funcdoc(func, docstr)
        if isinstance(docstr_func, six.string_types):
            olddoc = meta_util_six.get_funcdoc(func)
            if olddoc is None:
                olddoc = ''
            newdoc = olddoc + docstr_func
            meta_util_six.set_funcdoc(func, newdoc)
            return func
        else:
            preserved_func = preserve_sig(func, docstr_func)
            return preserved_func
    return docstr_applier


def preserve_sig(wrapper, orig_func, force=False):
    """
    Decorates a wrapper function.

    It seems impossible to presever signatures in python 2 without eval
    (Maybe another option is to write to a temporary module?)

    Args:
        wrapper: the function wrapping orig_func to change the signature of
        orig_func: the original function to take the signature from

    References:
        http://emptysqua.re/blog/copying-a-python-functions-signature/
        https://code.google.com/p/micheles/source/browse/decorator/src/decorator.py

    TODO:
        checkout funcsigs
        https://funcsigs.readthedocs.org/en/latest/

    CommandLine:
        python -m utool.util_decor --test-preserve_sig

    Example:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> #ut.rrrr(False)
        >>> def myfunction(self, listinput_, arg1, *args, **kwargs):
        >>>     " just a test function "
        >>>     return [x + 1 for x in listinput_]
        >>> #orig_func = ut.take
        >>> orig_func = myfunction
        >>> wrapper = ut.accepts_scalar_input2([0])(orig_func)
        >>> _wrp_preserve1 = ut.preserve_sig(wrapper, orig_func, True)
        >>> _wrp_preserve2 = ut.preserve_sig(wrapper, orig_func, False)
        >>> print('_wrp_preserve2 = %r' % (_wrp_preserve1,))
        >>> print('_wrp_preserve2 = %r' % (_wrp_preserve2,))
        >>> #print('source _wrp_preserve1 = %s' % (ut.get_func_sourcecode(_wrp_preserve1),))
        >>> #print('source _wrp_preserve2 = %s' % (ut.get_func_sourcecode(_wrp_preserve2)),)
        >>> result = str(_wrp_preserve1)
        >>> print(result)
    """
    #if True:
    #    import functools
    #    return functools.wraps(orig_func)(wrapper)
    from utool._internal import meta_util_six
    from utool import util_str
    from utool import util_inspect

    if wrapper is orig_func:
        # nothing to do
        return orig_func
    orig_docstr = meta_util_six.get_funcdoc(orig_func)
    orig_docstr = '' if orig_docstr is None else orig_docstr
    orig_argspec = util_inspect.get_func_argspec(orig_func)
    wrap_name = meta_util_six.get_funccode(wrapper).co_name
    orig_name = meta_util_six.get_funcname(orig_func)

    # At the very least preserve info in a dictionary
    _utinfo = {}
    _utinfo['orig_func'] = orig_func
    _utinfo['wrap_name'] = wrap_name
    _utinfo['orig_name'] = orig_name
    _utinfo['orig_argspec'] = orig_argspec

    if hasattr(wrapper, '_utinfo'):
        parent_wrapper_utinfo = wrapper._utinfo
        _utinfo['parent_wrapper_utinfo'] = parent_wrapper_utinfo
    if hasattr(orig_func, '_utinfo'):
        parent_orig_utinfo = orig_func._utinfo
        _utinfo['parent_orig_utinfo'] = parent_orig_utinfo

    # environment variable is set if you are building documentation
    # preserve sig if building docs
    building_docs = os.environ.get('UTOOL_AUTOGEN_SPHINX_RUNNING', 'OFF') == 'ON'

    if force or SIG_PRESERVE or building_docs:
        # PRESERVES ALL SIGNATURES WITH EXECS
        src_fmt = r'''
        def _wrp_preserve{defsig}:
            """ {orig_docstr} """
            try:
                return wrapper{callsig}
            except Exception as ex:
                import utool as ut
                msg = ('Failure in signature preserving wrapper:\n')
                ut.printex(ex, msg)
                raise
        '''
        # Put wrapped function into a scope
        globals_ =  {'wrapper': wrapper}
        locals_ = {}
        # argspec is :ArgSpec(args=['bar', 'baz'], varargs=None, keywords=None,
        # defaults=(True,))
        # get orig functions argspec
        # get functions signature
        # Get function call signature (no defaults)
        # Define an exec function
        argspec = inspect.getargspec(orig_func)
        (args, varargs, varkw, defaults) = argspec
        defsig = inspect.formatargspec(*argspec)
        callsig = inspect.formatargspec(*argspec[0:3])
        # TODO:
        # ut.func_defsig
        # ut.func_callsig
        src_fmtdict = dict(defsig=defsig, callsig=callsig, orig_docstr=orig_docstr)
        src = textwrap.dedent(src_fmt).format(**src_fmtdict)
        # Define the new function on the fly
        # (I wish there was a non exec / eval way to do this)
        #print(src)
        code = compile(src, '<string>', 'exec')
        six.exec_(code, globals_, locals_)
        #six.exec_(src, globals_, locals_)
        # Use functools.update_wapper to complete preservation
        _wrp_preserve = functools.update_wrapper(locals_['_wrp_preserve'], orig_func)
        # Keep debug info
        _utinfo['src'] = src
        # Set an internal sig variable that we may use
        #_wrp_preserve.__sig__ = defsig
    else:
        # PRESERVES SOME SIGNATURES NO EXEC
        # signature preservation is turned off. just preserve the name.
        # Does not use any exec or eval statments.
        _wrp_preserve = functools.update_wrapper(wrapper, orig_func)
        # Just do something to preserve signature

    DEBUG_WRAPPED_DOCSTRING = False
    if DEBUG_WRAPPED_DOCSTRING:
        new_docstr_fmtstr = util_str.codeblock(
            '''
            Wrapped function {wrap_name}({orig_name})

            orig_argspec = {orig_argspec}

            orig_docstr = {orig_docstr}
            '''
        )
    else:
        new_docstr_fmtstr = util_str.codeblock(
            '''
            {orig_docstr}
            '''
        )
    new_docstr = new_docstr_fmtstr.format(
        wrap_name=wrap_name, orig_name=orig_name, orig_docstr=orig_docstr,
        orig_argspec=orig_argspec)
    meta_util_six.set_funcdoc(_wrp_preserve, new_docstr)
    _wrp_preserve._utinfo = _utinfo
    return _wrp_preserve


def dummy_args_decor(*args, **kwargs):
    def dummy_args_closure(func):
        return func
    return dummy_args_closure


class classproperty(property):
    """
    Decorates a method turning it into a classattribute

    References:
        https://stackoverflow.com/questions/1697501/python-staticmethod-with-property
    """
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_decor; utool.doctest_funcs(utool.util_decor)"
        python -m utool.util_decor
        python -m utool.util_decor --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
NEEDS CLEANUP SO IT EITHER DOES THE IMPORTS OR GENERATES THE FILE

python -c "import utool"

"""
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import multiprocessing
import textwrap
#import types

#DEBUG_IMPORTS = '--debug-imports' in sys.argv
#----------
# EXECUTORS
#----------

def __excecute_imports(module, modname, imports, verbose=False):
    """ Module Imports """
    # level: -1 is a the Python2 import strategy
    # level:  0 is a the Python3 absolute import
    if verbose:
        print('[UTIL_IMPORT] EXECUTING %d IMPORT TUPLES' % (len(imports),))
    level = 0
    for name in imports:
        if level == -1:
            tmp = __import__(name, globals(), locals(), fromlist=[], level=level)
        elif level == 0:
            # FIXME: should support unicode. Maybe just a python2 thing
            tmp = __import__(modname, globals(), locals(), fromlist=[str(name)], level=level)


def __execute_fromimport(module, modname, import_tuples, verbose=False):
    """ Module From Imports """
    if verbose:
        print('[UTIL_IMPORT] EXECUTING %d FROM IMPORT TUPLES' % (len(import_tuples),))
    from_imports = __get_from_imports(import_tuples)
    for name, fromlist in from_imports:
        full_modname = '.'.join((modname, name))
        tmp = __import__(full_modname, globals(), locals(), fromlist=fromlist, level=0)
        for attrname in fromlist:
            setattr(module, attrname, getattr(tmp, attrname))
    return from_imports


def __execute_fromimport_star(module, modname, import_tuples, ignore_list=[],
                              ignore_startswith=[], ignore_endswith=[],
                              check_not_imported=True, verbose=False,
                              veryverbose=False):
    r"""
    Effectively import * statements

    The dynamic_import must happen before any * imports otherwise it wont catch
    anything.

    Ignore:
        ignore_startswith = []
        ignore_endswith = []
        check_not_imported = False
        verbose = True
        veryverbose = True
    """
    if verbose:
        print('[UTIL_IMPORT] EXECUTE %d FROMIMPORT STAR TUPLES.' % (len(import_tuples),))
    from_imports = []
    # Explicitly ignore these special functions (usually stdlib functions)
    ignoreset = set(['print', 'print_', 'printDBG', 'rrr', 'profile',
                     'print_function', 'absolute_import', 'division', 'zip',
                     'map', 'range', 'list', 'zip_longest', 'filter', 'filterfalse',
                     'dirname', 'realpath', 'join', 'exists', 'normpath',
                     'splitext', 'expanduser', 'relpath', 'isabs',
                     'commonprefix', 'basename', 'input', 'reduce',
                     #'OrderedDict',
                     #'product',
                     ] + ignore_list)
                     #'isdir', 'isfile', '

    #def is_defined_by_module2(item, module):
    #    belongs = False
    #    if hasattr(item, '__module__'):
    #        belongs = item.__module__ == module.__name__
    #    elif hasattr(item, 'func_globals'):
    #        belongs = item.func_globals['__name__'] == module.__name__
    #    return belongs

    for name, fromlist in import_tuples:
        #absname = modname + '.' + name
        child_module = sys.modules[modname + '.' + name]
        # Check if the variable already belongs to the module
        varset = set(vars(module)) if check_not_imported else set()
        fromset = set(fromlist) if fromlist is not None else set()
        def valid_attrname(attrname):
            """
            Guess if the attrname is valid based on its name
            """
            is_forced  = attrname in fromset
            is_private = attrname.startswith('_')
            is_conflit = attrname in varset
            is_module  = attrname in sys.modules  # Isn't fool proof (next step is)
            is_ignore1 = attrname in ignoreset
            is_ignore2 = any([attrname.startswith(prefix) for prefix in ignore_startswith])
            is_ignore3 = any([attrname.endswith(suffix) for suffix in ignore_endswith])
            is_ignore  = any((is_ignore1, is_ignore2, is_ignore3))
            is_valid = not any((is_ignore, is_private, is_conflit, is_module))
            #is_valid = is_valid and is_defined_by_module2(getattr(child_module, attrname), child_module)
            return (is_forced or is_valid)
        allattrs = dir(child_module)
        fromlist_ = [attrname for attrname in allattrs if valid_attrname(attrname)]
        #if verbose:
        #    print('[UTIL_IMPORT]     name=%r, len(allattrs)=%d' % (name, len(allattrs)))
        #if verbose:
        #    print('[UTIL_IMPORT]     name=%r, len(fromlist_)=%d' % (name, len(fromlist_)))
        valid_fromlist_ = []
        for attrname in fromlist_:
            attrval = getattr(child_module, attrname)
            try:
                # Disallow fromimport modules
                forced = attrname in fromset
                if not forced and getattr(attrval, '__name__') in sys.modules:
                    if veryverbose:
                        print('[UTIL_IMPORT] not importing: %r' % attrname)
                    continue
            except AttributeError:
                pass
            if veryverbose:
                print('[UTIL_IMPORT] %s is importing: %r' % (modname, attrname))
            valid_fromlist_.append(attrname)
            setattr(module, attrname, attrval)
        if verbose:
            print('[UTIL_IMPORT]     name=%r, len(valid_fromlist_)=%d' % (name, len(valid_fromlist_)))
        from_imports.append((name, valid_fromlist_))
    return from_imports

#----------
# PARSERS
#----------

def __get_from_imports(import_tuples):
    """ Returns import names and fromlist
    import_tuples are specified as
    (name, fromlist, ispackage)
    """
    from_imports = [(tup[0], tup[1]) for tup in import_tuples
                    if tup[1] is not None and len(tup[1]) > 0]
    return from_imports

#----------
# STRING MAKERS
#----------

def _initstr(modname, imports, from_imports, inject_execstr, withheader=True):
    """ Calls the other string makers """
    header         = _make_module_header() if withheader else ''
    import_str     = _make_imports_str(imports, modname)
    fromimport_str = _make_fromimport_str(from_imports, modname)
    initstr = '\n'.join([str_ for str_ in [
        header,
        import_str,
        fromimport_str,
        inject_execstr,
    ] if len(str_) > 0])
    return initstr

def _make_module_header():
    return '\n'.join([
        '# flake8: noqa',
        'from __future__ import absolute_import, division, print_function, unicode_literals'])

def _make_imports_str(imports, rootmodname='.'):
    imports_fmtstr = 'from {rootmodname} import %s'.format(rootmodname=rootmodname)
    return '\n'.join([imports_fmtstr % (name,) for name in imports])

def _make_fromimport_str(from_imports, rootmodname='.'):
    from utool import util_str
    if rootmodname == '.':
        # dot is already taken care of in fmtstr
        rootmodname = ''
    def _pack_fromimport(tup):
        name, fromlist = tup[0], tup[1]
        from_module_str = 'from {rootmodname}.{name} import ('.format(rootmodname=rootmodname, name=name)
        newline_prefix = (' ' * len(from_module_str))
        if len(fromlist) > 0:
            rawstr = from_module_str + ', '.join(fromlist) + ',)'
        else:
            rawstr = ''

        # not sure why this isn't 76? >= maybe?
        packstr = util_str.pack_into(rawstr, textwidth=75,
                                     newline_prefix=newline_prefix,
                                    break_words=False)
        return packstr
    from_str = '\n'.join(map(_pack_fromimport, from_imports))
    return from_str

def _inject_execstr(modname, import_tuples):
    """ Injection and Reload String Defs """
    if modname == 'utool':
        # Special case import of the util_inject module
        injecter = 'util_inject'
        injecter_import = ''
    else:
        # Normal case implicit import of util_inject
        injecter_import = 'import utool'
        injecter = 'utool'
    injectstr_fmt = textwrap.dedent(
        r'''
        # STARTBLOCK
        {injecter_import}
        print, rrr, profile = {injecter}.inject2(__name__, '[{modname}]')


        def reassign_submodule_attributes(verbose=1):
            """
            Updates attributes in the __init__ modules with updated attributes
            in the submodules.
            """
            import sys
            if verbose and '--quiet' not in sys.argv:
                print('dev reimport')
            # Self import
            import {modname}
            # Implicit reassignment.
            seen_ = set([])
            for tup in IMPORT_TUPLES:
                if len(tup) > 2 and tup[2]:
                    continue  # dont import package names
                submodname, fromimports = tup[0:2]
                submod = getattr({modname}, submodname)
                for attr in dir(submod):
                    if attr.startswith('_'):
                        continue
                    if attr in seen_:
                        # This just holds off bad behavior
                        # but it does mimic normal util_import behavior
                        # which is good
                        continue
                    seen_.add(attr)
                    setattr({modname}, attr, getattr(submod, attr))


        def reload_subs(verbose=1):
            """ Reloads {modname} and submodules """
            if verbose:
                print('Reloading {modname} submodules')
            rrr(verbose > 1)
            def wrap_fbrrr(mod):
                def fbrrr(*args, **kwargs):
                    """ fallback reload """
                    if verbose > 0:
                        print('Auto-reload (using rrr) not setup for mod=%r' % (mod,))
                return fbrrr
            def get_rrr(mod):
                if hasattr(mod, 'rrr'):
                    return mod.rrr
                else:
                    return wrap_fbrrr(mod)
            def get_reload_subs(mod):
                return getattr(mod, 'reload_subs', wrap_fbrrr(mod))
            {reload_body}
            rrr(verbose > 1)
            try:
                # hackish way of propogating up the new reloaded submodule attributes
                reassign_submodule_attributes(verbose=verbose)
            except Exception as ex:
                print(ex)
        rrrr = reload_subs
        # ENDBLOCK
        ''')
    injectstr_fmt = injectstr_fmt.replace('# STARTBLOCK', '')
    injectstr_fmt = injectstr_fmt.replace('# ENDBLOCK', '')
    rrrdir_fmt  = '    get_reload_subs({modname})(verbose=verbose)'
    rrrfile_fmt = '    get_rrr({modname})(verbose > 1)'

    def _reload_command(tup):
        if len(tup) > 2 and tup[2] is True:
            return rrrdir_fmt.format(modname=tup[0])
        else:
            return rrrfile_fmt.format(modname=tup[0])
    reload_body = '\n'.join(map(_reload_command, import_tuples)).strip()
    format_dict = {
        'modname': modname,
        'reload_body': reload_body,
        'injecter': injecter,
        'injecter_import': injecter_import,
    }
    inject_execstr = injectstr_fmt.format(**format_dict).strip()
    return inject_execstr

#----------
# PUBLIC FUNCTIONS
#----------

def dynamic_import(modname, import_tuples, developing=True, ignore_froms=[],
                   dump=False, ignore_startswith=[], ignore_endswith=[],
                   ignore_list=[], check_not_imported=True, return_initstr=False,
                   verbose=False):
    """
    MAIN ENTRY POINT

    Dynamically import listed util libraries and their attributes.
    Create reload_subs function.

    Using __import__ like this is typically not considered good style However,
    it is better than import * and this will generate the good file text that
    can be used when the module is 'frozen"

    Returns:
        str: init_inject_str - by default all imports are executed in this
            function and only the remainig code needed to be executed is
            returned to define the reload logic.

        str, str: init_inject_str, init_str - if return_initstr is True then
            also returns init_str defining the from imports.

    Ignore:
        ignore_startswith = []
        ignore_endswith = []
        check_not_imported = True
        verbose = True
    """
    if verbose:
        print('[UTIL_IMPORT] Running Dynamic Imports for modname=%r ' % modname)
    # Get the module that will be imported into
    try:
        module = sys.modules[modname]
    except:
        module = __import__(modname)
    # List of modules to be imported
    imports = [tup[0] for tup in import_tuples]
    # Import the modules
    __excecute_imports(module, modname, imports, verbose=verbose)
    # If developing do explicit import stars
    if developing:
        from_imports = __execute_fromimport_star(module, modname, import_tuples,
                                                 ignore_list=ignore_list,
                                                 ignore_startswith=ignore_startswith,
                                                 ignore_endswith=ignore_endswith,
                                                 check_not_imported=check_not_imported,
                                                 verbose=verbose)
    else:
        from_imports = __execute_fromimport(module, modname, import_tuples, verbose=verbose)

    inject_execstr = _inject_execstr(modname, import_tuples)

    # If requested: print what the __init__ module should look like
    dump_requested = (('--dump-%s-init' % modname) in sys.argv or
                      ('--print-%s-init' % modname) in sys.argv) or dump
    overwrite_requested = ('--update-%s-init' % modname) in sys.argv
    if verbose:
        print('[UTIL_IMPORT] Finished Dynamic Imports for modname=%r ' % modname)

    if dump_requested:
        is_main_proc = multiprocessing.current_process().name == 'MainProcess'
        if is_main_proc:
            from utool import util_str
            initstr = _initstr(modname, imports, from_imports, inject_execstr)
            print(util_str.indent(initstr))
    # Overwrite the __init__.py file with new explicit imports
    if overwrite_requested:
        """
        SeeAlso:
            util_inject.inject_python_code
            util_str.replace_between_tags
        """
        is_main_proc = multiprocessing.current_process().name == 'MainProcess'
        if is_main_proc:
            from utool import util_str
            from os.path import join, exists
            initstr = _initstr(modname, imports, from_imports, inject_execstr, withheader=False)
            new_else = util_str.indent(initstr)
            #print(new_else)
            # Get path to init file so we can overwrite it
            init_fpath = join(module.__path__[0], '__init__.py')
            print('attempting to update: %r' % init_fpath)
            assert exists(init_fpath)
            new_lines = []
            editing = False
            updated = False
            #start_tag = '# <AUTOGEN_INIT>'
            #end_tag = '# </AUTOGEN_INIT>'
            with open(init_fpath, 'r') as file_:
                #text = file_.read()
                lines = file_.readlines()
                for line in lines:
                    if not editing:
                        new_lines.append(line)
                    if line.strip().startswith('# <AUTOGEN_INIT>'):
                        new_lines.append('\n' + new_else + '\n    # </AUTOGEN_INIT>\n')
                        editing = True
                        updated = True
                    if line.strip().startswith('# </AUTOGEN_INIT>'):
                        editing = False
            # TODO:
            #new_text = util_str.replace_between_tags(text, new_else, start_tag, end_tag)
            if updated:
                print('writing updated file: %r' % init_fpath)
                new_text = ''.join(new_lines)
                with open(init_fpath, 'w') as file_:
                    file_.write(new_text)
            else:
                print('no write hook for file: %r' % init_fpath)
    if return_initstr:
        initstr = _initstr(modname, imports, from_imports, '', withheader=False)
        return inject_execstr, initstr
    else:
        return inject_execstr


def make_initstr(modname, import_tuples, verbose=False):
    """
    Just creates the string representation. Does no importing.
    """
    imports = [tup[0] for tup in import_tuples]
    from_imports = __get_from_imports(import_tuples)
    inject_execstr = _inject_execstr(modname, import_tuples)
    return _initstr(modname, imports, from_imports, inject_execstr)


def make_import_tuples(module_path, exclude_modnames=[]):
    """ Infer the import_tuples from a module_path """
    from utool import util_path
    kwargs = dict(private=False, full=False)
    module_list  = util_path.ls_modulefiles(module_path, noext=True, **kwargs)
    package_list = util_path.ls_moduledirs(module_path, **kwargs)
    exclude_set = set(exclude_modnames)
    module_import_tuples = [(modname, None) for modname in module_list
                            if modname not in exclude_set]
    package_import_tuples = [(modname, None, True)  for modname in package_list
                            if modname not in exclude_set]
    import_tuples = (module_import_tuples + package_import_tuples)
    return import_tuples
